#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Optional, List

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.item.ctx import CtxItem, CtxMeta


class BaseRenderer:
    def __init__(self, window=None):
        """
        Base renderer

        :param window: Window instance
        """
        self.window = window
        self.tab = None

    def set_tab(self, tab: Tab):
        """
        Append tab

        :param tab: Tab
        """
        self.tab = tab

    def prepare(self):
        """
        Prepare renderer
        """
        pass

    def get_pid(self, meta: CtxMeta):
        """
        Get PID for context meta
        
        :param meta: context PID
        """
        pass

    def is_stream(self) -> bool:
        """
        Check if it is a stream

        :return: True if stream is enabled
        """
        return self.window.core.config.get("stream")

    def begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render begin
        
        :param meta: context meta
        :param ctx: context item
        :param stream: True if stream
        """
        pass

    def end(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end
        
        :param meta: context meta
        :param ctx: context item
        :param stream: True if stream
        """
        pass

    def end_extra(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            stream: bool = False
    ):
        """
        Render end extra
        
        :param meta: context meta
        :param ctx: context item
        :param stream: True if stream
        """
        pass

    def stream_begin(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream begin
        
        :param meta: context meta
        :param ctx: context item
        """
        pass  # do nothing

    def stream_end(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Render stream end
        
        :param meta: context meta
        :param ctx: context item
        """
        pass  # do nothing    

    def clear_output(
            self,
            meta: Optional[CtxMeta] = None
    ):
        """
        Clear output

        :param meta: context meta
        """
        pass

    def clear_input(self):
        """Clear input"""
        pass

    def reset(self, meta: CtxMeta = None):
        """
        Reset

        :param meta: context meta
        """
        pass

    def reload(self):
        """Reload all outputs, called externally only on theme change to redraw content"""
        pass

    def append_context(
            self,
            meta: CtxMeta,
            items: List[CtxItem],
            clear: bool = True
    ):
        """
        Append all context to output
        
        :param meta: context meta
        :param items: context items
        :param clear: True if clear all output before append
        """
        pass

    def append_input(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            flush: bool = True,
            append: bool = False
    ):
        """
        Append text input to output
        
        :param meta: context meta
        :param ctx: context item
        :param flush: True if flush
        :param append: True to force append node
        """
        pass

    def append_output(
            self,
            meta: CtxMeta,
            ctx: CtxItem
    ):
        """
        Append text output to output
        
        :param meta: context meta
        :param ctx: context item
        """
        pass

    def append_extra(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            footer: bool = False
    ):
        """
        Append extra data (images, files, etc.) to output
        
        :param meta: context meta
        :param ctx: context item
        :param footer: True if it is a footer
        """
        pass

    def append_chunk(
            self,
            meta: CtxMeta,
            ctx: CtxItem,
            text_chunk: str,
            begin: bool = False
    ):
        """
        Append output chunk to output
        
        :param meta: context meta
        :param ctx: context item
        :param text_chunk: text chunk
        :param begin: if it is the beginning of the text
        """
        pass

    def remove_item(self, id: int):
        """
        Remove item from output

        :param id: context item ID
        """
        pass

    def remove_items_from(self, id: int):
        """
        Remove item from output

        :param id: context item ID
        """
        pass

    def on_edit_submit(self, ctx: CtxItem):
        """
        On edit submit

        :param ctx: context item
        """
        self.window.controller.ctx.refresh(restore_model=False)  # allow model change

    def on_remove_submit(self, ctx: CtxItem):
        """
        On remove submit

        :param ctx: context item
        """
        pass

    def on_reply_submit(self, ctx: CtxItem):
        """
        On regenerate submit

        :param ctx: context item
        """
        self.window.controller.ctx.refresh(restore_model=False)  # allow model change

    def clear_all(self):
        """Clear all"""
        pass

    def on_load(self, meta: CtxMeta = None):
        """
        On load (meta)

        :param meta: context metam
        """
        pass

    def on_page_loaded(
            self,
            meta: Optional[CtxMeta] = None,
            tab: Optional[Tab] = None):
        """
        On page loaded callback

        :param meta: context meta
        :param tab: Tab
        """
        self.tab = tab

    def on_enable_edit(
            self,
            live: bool = True
    ):
        """
        On enable edit icons

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_disable_edit(
            self,
            live: bool = True
    ):
        """
        On disable edit icons

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_enable_timestamp(
            self,
            live: bool = True
    ):
        """
        On enable timestamp

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_disable_timestamp(
            self,
            live: bool = True
    ):
        """
        On disable timestamp

        :param live: True if live update
        """
        if live:  # default behavior
            self.window.controller.ctx.refresh()

    def on_theme_change(self):
        """On theme change"""
        pass

    def get_scroll_position(self) -> int:
        """
        Get scroll position

        :return: scroll position
        """
        return 0

    def set_scroll_position(self, position: int):
        """
        Set scroll position

        :param position: scroll position
        """
        pass

    def tool_output_append(
            self,
            meta:
            CtxMeta,
            content: str
    ):
        """
        Add tool output (append)

        :param meta: context meta
        :param content: content
        """
        pass

    def tool_output_update(
            self,
            meta: CtxMeta,
            content: str
    ):
        """
        Replace tool output

        :param meta: context meta
        :param content: content
        """
        pass

    def tool_output_clear(
            self,
            meta: CtxMeta
    ):
        """
        Clear tool output

        :param meta: context meta
        """
        pass

    def tool_output_begin(
            self,
            meta: CtxMeta
    ):
        """
        Begin tool output

        :param meta: context meta
        """
        pass

    def tool_output_end(self):
        """End tool output"""
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.16 01:00:00                  #
# ================================================== #

import json
from typing import Dict, Tuple, Any, Optional

from PySide6.QtCore import Qt
from PySide6.QtWidgets import QVBoxLayout, QLabel, QHBoxLayout, QWidget

from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.ui.widget.option.input import OptionInput


class Loaders:
    def __init__(self, window=None):
        """
        UI - loaders components

        :param window: Window instance
        """
        self.window = window

    def handle_options(
            self,
            select_loader,
            prefix_options,
            prefix_config
    ) -> Tuple[bool, Optional[str], Dict[str, Any], Dict[str, Any]]:
        """
        Handle options

        :param select_loader: loader selection
        :param prefix_options: prefix for options
        :param prefix_config: prefix for config
        :return: bool, loader name, input_params, input_config
        """
        input_params = {}
        input_config = {}
        loader = select_loader.get_value()
        if not loader:
            return False, loader, input_params, input_config
        loaders = self.window.core.idx.indexing.get_external_instructions()
        if loader in loaders:
            params = loaders[loader]
            for k in params["args"]:
                key_path = prefix_options + "." + loader + "." + k
                if key_path in self.window.ui.nodes:
                    tmp_value = self.window.ui.nodes[key_path].text()
                    type = params["args"][k]["type"]
                    try:
                        if tmp_value:
                            if type == "int":
                                tmp_value = int(tmp_value)
                            elif type == "float":
                                tmp_value = float(tmp_value)
                            elif type == "bool":
                                if tmp_value.lower() in ["true", "1"]:
                                    tmp_value = True
                                else:
                                    tmp_value = False
                            elif type == "list":
                                tmp_value = tmp_value.split(",")
                            elif type == "dict":
                                tmp_value = json.loads(tmp_value)
                            input_params[k] = tmp_value
                    except Exception as e:
                        self.window.core.debug.log(e)
                        self.window.ui.dialogs.alert(e)

        loaders = self.window.core.idx.indexing.get_external_config()
        if loader in loaders:
            params = loaders[loader]
            for k in params:
                key_path = prefix_config + "." + loader + "." + k
                type = params[k]["type"]
                if key_path in self.window.ui.nodes:
                    tmp_value = self.window.ui.nodes[key_path].text()
                    try:
                        if tmp_value:
                            if type == "int":
                                tmp_value = int(tmp_value)
                            elif type == "float":
                                tmp_value = float(tmp_value)
                            elif type == "bool":
                                if tmp_value.lower() in ["true", "1"]:
                                    tmp_value = True
                                else:
                                    tmp_value = False
                            elif type == "list":
                                tmp_value = tmp_value.split(",")
                            elif type == "dict":
                                tmp_value = json.loads(tmp_value)
                            input_config[k] = tmp_value
                    except Exception as e:
                        self.window.core.debug.log(e)
                        self.window.ui.dialogs.alert(e)

        return True, loader, input_params, input_config

    def setup_loader_options(self):
        """Setup loader options"""
        inputs = {}
        groups = {}
        loaders = self.window.core.idx.indexing.get_external_instructions()
        for loader in loaders:
            params = loaders[loader]
            inputs[loader] = {}
            group = QVBoxLayout()
            for k in params["args"]:
                label = k
                description = None
                is_label = False
                if "label" in params["args"][k]:
                    label = params["args"][k]["label"]
                    is_label = True
                if "description" in params["args"][k]:
                    description = params["args"][k]["description"]
                option_id = "web.loader." + loader + ".option." + k
                option_widget = OptionInput(self.window, "tool.indexer", option_id, {
                    "label": label,
                    "value": "",
                })
                option_widget.setPlaceholderText(params["args"][k]["type"])
                inputs[loader][k] = option_widget

                option_label = QLabel(label)
                option_label.setToolTip(k)

                row = QHBoxLayout()  # cols
                row.addWidget(option_label)
                row.addWidget(option_widget)
                row.setContentsMargins(5, 0, 5, 0)

                option_layout = QVBoxLayout()
                option_layout.addLayout(row)
                if description:
                    option_layout.addWidget(HelpLabel(description))

                option_layout.setContentsMargins(5, 0, 0, 0)

                group.addLayout(option_layout)
                group.setContentsMargins(0, 0, 0, 0)

            group_widget = QWidget()
            group_widget.setLayout(group)
            groups[loader] = group_widget

        return inputs, groups

    def setup_loader_config(self):
        """Setup loader config"""
        inputs = {}
        groups = {}
        loaders = self.window.core.idx.indexing.get_external_config()
        for loader in loaders:
            params = loaders[loader]
            inputs[loader] = {}
            group = QVBoxLayout()
            for k in params:
                label = k
                description = None
                is_label = False
                if "label" in params[k]:
                    label = params[k]["label"]
                    is_label = True
                if "description" in params[k]:
                    description = params[k]["description"]
                option_id = "web.loader." + loader + ".config." + k
                option_widget = OptionInput(self.window, "tool.indexer", option_id, {
                    "label": label,
                    "value": params[k]["value"],
                })
                try:
                    if params[k]["value"] is not None:
                        if params[k]["type"] == "list" and isinstance(params[k]["value"], list):
                            option_widget.setText(", ".join(params[k]["value"]))
                        elif params[k]["type"] == "dict" and isinstance(params[k]["value"], dict):
                            option_widget.setText(json.dumps(params[k]["value"]))
                        else:
                            option_widget.setText(str(params[k]["value"]))
                except Exception as e:
                    self.window.core.debug.log(e)

                option_widget.setPlaceholderText(params[k]["type"])
                inputs[loader][k] = option_widget

                option_label = QLabel(label)
                option_label.setToolTip(k)

                row = QHBoxLayout()  # cols
                row.addWidget(option_label)
                row.addWidget(option_widget)
                row.setContentsMargins(5, 0, 5, 0)

                option_layout = QVBoxLayout()
                option_layout.addLayout(row)
                if description:
                    option_layout.addWidget(HelpLabel(description))

                option_layout.setContentsMargins(5, 0, 0, 0)

                group.addLayout(option_layout)
                group.setContentsMargins(0, 0, 0, 0)

            group_widget = QWidget()
            group_widget.setLayout(group)
            groups[loader] = group_widget

        return inputs, groups


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.02.27 18:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QTimer, QRect, Signal
from PySide6.QtGui import QCursor, QAction, QIcon
from PySide6.QtWidgets import QLabel, QLineEdit, QToolTip
import webbrowser

from pygpt_net.utils import trans
import pygpt_net.icons_rc


class HelpLabel(QLabel):
    def __init__(self, text, window=None):
        super().__init__(text, window)
        self.window = window
        self.setWordWrap(True)
        self.setContentsMargins(3, 3, 3, 3)
        self.setProperty('class', 'label-help')


class TitleLabel(QLabel):
    def __init__(self, text, window=None):
        super().__init__(text, window)
        self.window = window
        self.setProperty('class', 'label-title')


class ChatStatusLabel(QLabel):
    def __init__(self, text, window=None):
        super().__init__(text, window)
        self.window = window
        self.setWordWrap(True)
        self.setAlignment(Qt.AlignRight)
        self.setProperty('class', 'label-chat-status')


class UrlLabel(QLabel):
    def __init__(self, text, url, parent=None):
        super().__init__(text, parent)
        self.text = text
        self.url = url
        self.update_url()
        self.setCursor(QCursor(Qt.PointingHandCursor))
        self.setToolTip(url)
        # self.setWordWrap(True)

    def update_url(self):
        text = self.text
        if text is not None and len(text) > 0:
            text += ': '
        text += self.url
        self.setText(text)

    def mousePressEvent(self, event):
        webbrowser.open(self.url)


class CmdLabel(QLineEdit):
    def __init__(self, window=None, text: str = ""):
        """
        Command label

        :param window: main window
        :param text: text
        """
        super(CmdLabel, self).__init__(window)
        self.window = window
        self.setStyleSheet(
            "font-weight: bold;"
            "font-size: 14px;"
            "padding: 5px;"
            "background: #020202;"
            "color: #fff;"
            "text-align: center;"
        )
        self.setText(text)
        self.setAlignment(Qt.AlignCenter)
        self.setReadOnly(True)
        self.action = QAction(self)
        self.action.setIcon(QIcon(":/icons/copy.svg"))
        self.action.triggered.connect(self.copy_to_clipboard)
        self.addAction(self.action, QLineEdit.TrailingPosition)

    def copy_to_clipboard(self):
        """Copy to clipboard"""
        self.selectAll()
        self.copy()
        self.deselect()
        self.action.setIcon(QIcon(":/icons/check.svg"))
        QToolTip.showText(QCursor.pos(), trans("clipboard.copied"), self, QRect(), 2000)
        timer = QTimer()
        timer.singleShot(2000, self.reset_icon)

    def reset_icon(self):
        """Reset icon"""
        self.action.setIcon(QIcon(":/icons/copy.svg"))


class IconLabel(QLabel):
    clicked = Signal()

    def __init__(self, icon: str, window=None):
        super().__init__("", window)
        self.window = window
        self.setWordWrap(True)
        self.setAlignment(Qt.AlignRight)
        self.setProperty('class', 'label-chat-status')
        self.setContentsMargins(0, 0, 0, 0)
        self.set_icon(icon)

    def set_icon(self, icon: str):
        self.setPixmap(QIcon(icon).pixmap(16, 16))

    def mousePressEvent(self, event):
        self.clicked.emit()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.15 04:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot, Signal

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    output = Signal(object, str)
    html_output = Signal(object)
    ipython_output = Signal(object)
    build_finished = Signal()
    clear = Signal()


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        for item in self.cmds:
            if self.is_stopped():
                break
            try:
                response = None
                if (item["cmd"] in self.plugin.allowed_cmds
                        and (self.plugin.has_cmd(item["cmd"]) or 'force' in item)):

                    if item["cmd"] == "send_email":
                        response = self.cmd_send_mail(item)

                    elif item["cmd"] == "get_emails":
                        response = self.cmd_receive_emails(item)

                    elif item["cmd"] == "get_email_body":
                        response = self.cmd_get_email_body(item)

                    if response:
                        responses.append(response)

            except Exception as e:
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

    def cmd_send_mail(self, item: dict) -> dict:
        """
        Send email command

        :param item: command item
        :return: response item
        """
        request = self.from_request(item)
        try:
            result = self.plugin.runner.smtp_send_email(
                ctx=self.ctx,
                item=item,
                request=request,
            )
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_receive_emails(self, item: dict) -> dict:
        """
        Receive emails command

        :param item: command item
        :return: response item
        """
        request = self.from_request(item)
        try:
            result = self.plugin.runner.smtp_receive_emails(
                ctx=self.ctx,
                item=item,
                request=request,
            )
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_get_email_body(self, item: dict) -> dict:
        """
        Get email body command

        :param item: command item
        :return: response item
        """
        request = self.from_request(item)
        try:
            result = self.plugin.runner.smtp_get_email_body(
                ctx=self.ctx,
                item=item,
                request=request,
            )
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import datetime
import math
import os
from typing import Optional, Union

import mss
import mss.tools
from PIL import Image, ImageDraw
from pynput.mouse import Controller

from PySide6.QtGui import QImage

from pygpt_net.core.events import KernelEvent
from pygpt_net.utils import trans


class Capture:
    def __init__(self, window=None):
        """
        Painter capture controller

        :param window: Window instance
        """
        self.window = window

    def camera(self):
        """Get image from camera"""
        if not self.window.controller.camera.is_enabled():
            self.window.controller.camera.enable_capture()
            self.window.controller.camera.setup_ui()
        frame = self.window.controller.camera.get_current_frame(False)
        if frame is None:
            return
        height, width, channel = frame.shape
        bytes = 3 * width
        image = QImage(frame.data, width, height, bytes, QImage.Format_RGB888)
        self.window.ui.painter.set_image(image)

    def capture_screen_with_custom_cursor(self, save_path: str) -> str:
        """
        Capture screen with custom cursor

        :param save_path: Save path
        :return: Save path
        """
        cursor_path = os.path.join(self.window.core.config.get_app_path(), "data", "icons", "cursor.png")

        with mss.mss() as sct:
            monitor = sct.monitors[1]
            screenshot = sct.grab(monitor)
            img = Image.frombytes('RGB', screenshot.size, screenshot.rgb)

        mouse = Controller()
        cursor_x, cursor_y = mouse.position

        cursor_x -= monitor['left']
        cursor_y -= monitor['top']

        img = img.convert('RGBA')

        cursor_img = Image.open(cursor_path).convert('RGBA')
        cursor_width, cursor_height = cursor_img.size

        paste_x = int(cursor_x - 20)
        paste_y = int(cursor_y - 20)

        img.paste(cursor_img, (paste_x, paste_y), cursor_img)

        img.save(save_path)
        return save_path

    def screenshot(
            self,
            attach_cursor: bool = False,
            silent: bool = False
    ) -> Optional[Union[str, bool]]:
        """
        Make screenshot and append to attachments

        :param attach_cursor: True to with custom cursor
        :param silent: Silent mode
        :return: Path to screenshot or False if failed
        """
        if not silent:
            # switch to vision mode if needed
            self.window.controller.chat.vision.switch_to_vision()

            # clear attachments before capture if needed
            if self.window.controller.attachment.is_capture_clear():
                self.window.controller.attachment.clear(True, auto=True)

        try:
            # prepare filename
            now = datetime.datetime.now()
            dt = now.strftime("%Y-%m-%d_%H-%M-%S")
            name = 'cap-' + dt
            path = os.path.join(self.window.controller.painter.common.get_capture_dir(), name + '.png')

            # capture screenshot
            if attach_cursor:
                if not self.capture_screen_with_custom_cursor(path):  # capture with custom cursor
                    return False
            else:
                with mss.mss(with_cursor=False) as sct:
                    monitor = sct.monitors[1]
                    sct_img = sct.grab(monitor)
                    mss.tools.to_png(sct_img.rgb, sct_img.size, output=path)

            self.attach(name, path, 'screenshot', silent=silent)

            if not silent:
                self.window.controller.painter.open(path)
                # show last capture time in status
                dt_info = now.strftime("%Y-%m-%d %H:%M:%S")
                event = KernelEvent(KernelEvent.STATUS, {
                    'status': trans("painter.capture.manual.captured.success") + ' ' + dt_info,
                })
                self.window.dispatch(event)
            return path

        except Exception as e:
            print("Screenshot capture exception", e)
            self.window.core.debug.log(e)

    def use(self):
        """Use current image"""
        # switch to vision mode if needed
        self.window.controller.chat.vision.switch_to_vision()

        # clear attachments before capture if needed
        if self.window.controller.attachment.is_capture_clear():
            self.window.controller.attachment.clear(True, auto=True)

        try:
            # prepare filename
            now = datetime.datetime.now()
            dt = now.strftime("%Y-%m-%d_%H-%M-%S")
            name = 'cap-' + dt
            path = os.path.join(self.window.controller.painter.common.get_capture_dir(), name + '.png')

            # capture
            self.window.ui.painter.image.save(path)
            self.attach(name, path)

            # show last capture time in status
            dt_info = now.strftime("%Y-%m-%d %H:%M:%S")
            event = KernelEvent(KernelEvent.STATUS, {
                'status': trans("painter.capture.manual.captured.success") + ' ' + dt_info,
            })
            self.window.dispatch(event)
            return True

        except Exception as e:
            print("Image capture exception", e)
            self.window.core.debug.log(e)

    def attach(
            self,
            name: str,
            path: str,
            type: str = 'drawing',
            silent: bool = False
    ):
        """
        Attach image to attachments

        :param name: image name
        :param path: image path
        :param type: capture type (drawing, screenshot)
        :param silent: silent mode
        """
        mode = self.window.core.config.get('mode')
        if type == 'drawing':
            title = trans('painter.capture.name.prefix') + ' ' + name
        elif type == 'screenshot':
            title = trans('screenshot.capture.name.prefix') + ' ' + name
        else:
            title = name
        title = title.replace('cap-', '').replace('_', ' ')

        # make attachment
        self.window.core.attachments.new(mode, title, path, False)
        self.window.core.attachments.save()

        if not silent:
            self.window.controller.attachment.update()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

from typing import Dict, List

from pygpt_net.provider.core.mode.json_file import JsonFileProvider
from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_IMAGE,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
)


class Modes:

    def __init__(self, window=None):
        """
        Modes core

        :param window: Window instance
        """
        self.window = window
        self.provider = JsonFileProvider(window)
        self.initialized = False
        self.all = [
            MODE_AGENT,
            MODE_AGENT_LLAMA,
            MODE_ASSISTANT,
            MODE_AUDIO,
            MODE_CHAT,
            MODE_COMPLETION,
            MODE_EXPERT,
            MODE_IMAGE,
            MODE_LANGCHAIN,
            MODE_LLAMA_INDEX,
            MODE_VISION,
            MODE_RESEARCH,
        ]
        self.items = {}

    def get_by_idx(self, idx) -> str:
        """
        Return mode by index

        :param idx: index of mode
        :return: mode name
        """
        modes = self.get_all()
        return list(modes.keys())[idx]

    def get_idx_by_name(self, name) -> int:
        """
        Return mode index by name

        :param name: mode name
        :return: mode index
        """
        modes = self.get_all()
        return list(modes.keys()).index(name)

    def get_all(self) -> Dict[str, List[str]]:
        """
        Return modes

        :return: Dict with modes
        """
        return self.items

    def get_default(self) -> str:
        """
        Return default mode name

        :return: default mode name
        """
        for id in self.items:
            if self.items[id].default:
                return id

    def get_next(self, mode: str) -> str:
        """
        Return next mode

        :param mode: current mode
        :return: next mode
        """
        modes = self.get_all()
        keys = list(modes.keys())
        idx = keys.index(mode)
        if idx + 1 < len(keys):
            return keys[idx + 1]
        return keys[0]

    def get_prev(self, mode: str) -> str:
        """
        Return previous mode

        :param mode: current mode
        :return: previous mode
        """
        modes = self.get_all()
        keys = list(modes.keys())
        idx = keys.index(mode)
        if idx - 1 >= 0:
            return keys[idx - 1]
        return keys[-1]

    def load(self):
        """Load modes"""
        self.items = self.provider.load()

    def save(self):
        """Save modes"""
        self.provider.save(self.items)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import List

from pygpt_net.item.ctx import CtxItem


class Bag:
    def __init__(self, window=None):
        """
        Context bag
        """
        self.window = window
        self.meta = None  # current meta
        self.tab_id = 0
        self.items = []

    def get_items(self) -> List[CtxItem]:
        """
        Return ctx items

        :return: ctx items
        """
        return self.items

    def set_items(self, items: List[CtxItem]):
        """
        Set ctx items

        :param items: ctx items
        """
        self.items = items

    def clear_items(self):
        """Clear items"""
        self.items = []

    def count_items(self) -> int:
        """
        Count ctx items

        :return: items count
        """
        return len(self.items)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 07:00:00                  #
# ================================================== #

from PySide6.QtWidgets import QTabWidget, QMenu, QPushButton
from PySide6.QtCore import Qt, Slot
from PySide6.QtGui import QAction, QIcon

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class AddButton(QPushButton):
    def __init__(self, window=None, column=None, tabs=None):
        super(AddButton, self).__init__(QIcon(":/icons/add.svg"), "", window)
        self.window = window
        self.column = column
        self.tabs = tabs
        self.setFixedSize(30, 25)
        self.setFlat(True)
        self.clicked.connect(
            lambda: self.window.controller.ui.tabs.new_tab(self.column.get_idx())
        )
        self.setObjectName('tab-add')
        self.setProperty('tabAdd', True)
        self.setToolTip(trans('action.tab.add.chat'))

    def mousePressEvent(self, event):
        """
        Mouse press event

        :param event: event
        """
        if event.button() == Qt.RightButton:
            idx = 0
            column_idx = self.column.get_idx()
            self.show_menu(idx, column_idx, event.globalPos())
        super(AddButton, self).mousePressEvent(event)

    def show_menu(self, index: int, column_idx: int, global_pos):
        """
        Show context menu

        :param index: index
        :param column_idx: column index
        :param global_pos: global position
        """
        context_menu = self.prepare_menu(index, column_idx)
        context_menu.exec(global_pos)

    def prepare_menu(self, index: int, column_idx: int) -> QMenu:
        """
        Prepare and return context menu

        :param index: index
        :param column_idx: column index
        :return: menu
        """
        menu = QMenu(self)

        actions = {}
        actions['add_chat'] = QAction(QIcon(":/icons/add.svg"), trans('action.tab.add.chat'), self)
        actions['add_chat'].triggered.connect(
            lambda: self.tabs.add_tab(index, column_idx, Tab.TAB_CHAT)
        )
        actions['add_notepad'] = QAction(QIcon(":/icons/add.svg"), trans('action.tab.add.notepad'), self)
        actions['add_notepad'].triggered.connect(
            lambda: self.tabs.add_tab(index, column_idx, Tab.TAB_NOTEPAD)
        )

        # add chat, add notepad
        menu.addAction(actions['add_chat'])
        menu.addAction(actions['add_notepad'])

        # add tools submenu
        self.window.controller.tools.append_tab_menu(self, menu, index, column_idx)

        return menu

class OutputTabs(QTabWidget):
    def __init__(self, window=None, column=None):
        super(OutputTabs, self).__init__(window)
        self.window = window
        self.active = True
        self.column = column
        self.setMinimumHeight(1)
        self.owner = None
        self.setMovable(True)
        self.init()

    def set_active(self, active: bool):
        """Set the active state of the tab bar."""
        self.active = active
        if self.active:
            self.setStyleSheet("QTabBar::tab { border-bottom-width: 2px; }")
        else:
            self.setStyleSheet("QTabBar::tab { border-bottom-width: 0px; }")

    def init(self):
        """Initialize"""
        # create the [+] button
        add_button = AddButton(self.window, self.column, self)

        # add the button to the top right corner of the tab bar
        self.setCornerWidget(add_button, corner=Qt.TopRightCorner)

        # connect signals
        self.currentChanged.connect(
            lambda: self.window.controller.ui.tabs.on_tab_changed(self.currentIndex(), self.column.get_idx())
        )
        self.tabBarClicked.connect(
            lambda: self.window.controller.ui.tabs.on_tab_clicked(self.currentIndex(), self.column.get_idx())
        )
        self.tabBarDoubleClicked.connect(
            lambda: self.window.controller.ui.tabs.on_tab_dbl_clicked(self.currentIndex(), self.column.get_idx())
        )
        self.tabCloseRequested.connect(
            lambda: self.window.controller.ui.tabs.on_tab_closed(self.currentIndex(), self.column.get_idx())
        )
        self.tabBar().tabMoved.connect(
            lambda: self.window.controller.ui.tabs.on_tab_moved(self.currentIndex(), self.column.get_idx())
        )

    def get_column(self):
        """
        Get column

        :return: OutputColumn
        """
        return self.column

    def setOwner(self, owner: Tab):
        """
        Set internal tab instance

        :param owner: parent tab instance
        """
        self.owner = owner

    def mousePressEvent(self, event):
        """
        Mouse press event

        :param event: event
        """
        if event.button() == Qt.RightButton:
            idx = self.tabBar().tabAt(event.pos())
            column_idx = self.column.get_idx()
            tab = self.window.core.tabs.get_tab_by_index(idx, column_idx)
            if tab is not None:
                if tab.type == Tab.TAB_NOTEPAD:
                    self.show_notepad_menu(idx, column_idx, event.globalPos())  # notepad
                elif tab.type == Tab.TAB_CHAT:
                    self.show_chat_menu(idx, column_idx, event.globalPos())  # chat
                elif tab.type == Tab.TAB_FILES:
                    self.show_files_menu(idx, column_idx, event.globalPos())  # files
                elif tab.type == Tab.TAB_TOOL:
                    self.show_tool_menu(idx, column_idx, event.globalPos())  # tool
                else:
                    self.show_default_menu(idx, column_idx, event.globalPos()) # default
        super(OutputTabs, self).mousePressEvent(event)

    def prepare_menu(self, index: int, column_idx: int) -> QMenu:
        """
        Prepare and return context menu

        :param index: index
        :param column_idx: column index
        :return: menu
        """
        menu = QMenu(self)

        actions = {}
        actions['add_chat'] = QAction(QIcon(":/icons/add.svg"), trans('action.tab.add.chat'), self)
        actions['add_chat'].triggered.connect(
            lambda: self.add_tab(index, column_idx, Tab.TAB_CHAT)
        )
        actions['add_notepad'] = QAction(QIcon(":/icons/add.svg"), trans('action.tab.add.notepad'), self)
        actions['add_notepad'].triggered.connect(
            lambda: self.add_tab(index, column_idx, Tab.TAB_NOTEPAD)
        )
        actions['edit'] = QAction(QIcon(":/icons/edit.svg"), trans('action.rename'), self)
        actions['edit'].triggered.connect(
            lambda: self.rename_tab(index, column_idx)
        )
        actions['move_right'] = QAction(QIcon(":/icons/forward"), trans('action.tab.move.right'), self)
        actions['move_right'].triggered.connect(
            lambda: self.window.controller.ui.tabs.move_tab(index, column_idx, 1)
        )
        actions['move_left'] = QAction(QIcon(":/icons/back"), trans('action.tab.move.left'), self)
        actions['move_left'].triggered.connect(
            lambda: self.window.controller.ui.tabs.move_tab(index, column_idx, 0)
        )

        # add chat, add notepad
        menu.addAction(actions['add_chat'])
        menu.addAction(actions['add_notepad'])

        # add tools submenu
        self.window.controller.tools.append_tab_menu(self, menu, index, column_idx)

        # rename tab
        menu.addAction(actions['edit'])

        # move tab left, move tab right
        if column_idx != 0:
            menu.addAction(actions['move_left'])
        if column_idx != 1:
            menu.addAction(actions['move_right'])

        return menu

    def show_notepad_menu(self, index: int, column_idx: int, global_pos):
        """
        Show notepad menu

        :param index: index
        :param column_idx: column index
        :param global_pos: global position
        """
        context_menu = self.prepare_menu(index, column_idx)
        actions = {}
        actions['close'] = QAction(QIcon(":/icons/close.svg"), trans('action.tab.close'), self)
        actions['close'].triggered.connect(
            lambda: self.close_tab(index, column_idx)
        )
        actions['close_all'] = QAction(QIcon(":/icons/close.svg"), trans('action.tab.close_all.notepad'), self)
        actions['close_all'].triggered.connect(
            lambda: self.close_all(Tab.TAB_NOTEPAD, column_idx)
        )
        context_menu.addAction(actions['close'])

        if self.window.core.tabs.count_by_type(Tab.TAB_NOTEPAD) > 1:
            context_menu.addAction(actions['close_all'])

        context_menu.exec(global_pos)

    def show_chat_menu(self, index: int, column_idx: int, global_pos):
        """
        Show chat menu

        :param index: index
        :param column_idx: column index
        :param global_pos: global position
        """
        context_menu = self.prepare_menu(index, column_idx)
        actions = {}
        actions['close'] = QAction(QIcon(":/icons/close.svg"), trans('action.tab.close'), self)
        actions['close'].triggered.connect(
            lambda: self.close_tab(index, column_idx)
        )
        actions['close_all'] = QAction(QIcon(":/icons/close.svg"), trans('action.tab.close_all.chat'), self)
        actions['close_all'].triggered.connect(
            lambda: self.close_all(Tab.TAB_CHAT, column_idx)
        )

        # at least one chat tab must be open
        if self.window.core.tabs.count_by_type(Tab.TAB_CHAT) > 1:
            context_menu.addAction(actions['close'])
            context_menu.addAction(actions['close_all'])

        context_menu.exec(global_pos)

    def show_files_menu(self, index: int, column_idx: int, global_pos):
        """
        Show files menu

        :param index: index
        :param column_idx: column index
        :param global_pos: global position
        """
        context_menu = self.prepare_menu(index, column_idx)
        actions = {}
        actions['refresh'] = QAction(QIcon(":/icons/reload.svg"), trans('action.refresh'), self)
        actions['refresh'].triggered.connect(
            lambda: self.window.controller.files.update_explorer()
        )
        context_menu.addAction(actions['refresh'])
        context_menu.exec(global_pos)

    def show_tool_menu(self, index: int, column_idx: int, global_pos):
        """
        Show tool menu

        :param index: index
        :param column_idx: column index
        :param global_pos: global position
        """
        context_menu = self.prepare_menu(index, column_idx)
        actions = {}
        actions['close'] = QAction(QIcon(":/icons/close.svg"), trans('action.tab.close'), self)
        actions['close'].triggered.connect(
            lambda: self.close_tab(index, column_idx)
        )
        context_menu.addAction(actions['close'])
        context_menu.exec(global_pos)

    def show_default_menu(self, index: int, column_idx: int, global_pos):
        """
        Show default menu

        :param index: index
        :param column_idx: column index
        :param global_pos: global position
        """
        context_menu = self.prepare_menu(index, column_idx)
        context_menu.exec(global_pos)

    @Slot()
    def rename_tab(self, index: int, column_idx: int):
        """
        Rename tab

        :param index: index
        :param column_idx: column index
        """
        self.window.controller.ui.tabs.rename(index, column_idx)

    @Slot()
    def close_tab(self, index: int, column_idx: int):
        """
        Close tab

        :param index: index
        :param column_idx: column index
        """
        self.window.controller.ui.tabs.close(index, column_idx)

    @Slot()
    def close_all(self, type, column_idx: int):
        """
        Close all tabs

        :param type: type
        :param column_idx: column index
        """
        self.window.controller.ui.tabs.close_all(type, column_idx)

    @Slot()
    def add_tab(self, index: int, column_idx: int, type: int, tool_id: str = None):
        """
        Add a new tab

        :param index: index
        :param column_idx: column index
        :param type: type
        :param tool_id: tool id
        """
        self.window.controller.ui.tabs.append(
            type=type,
            tool_id=tool_id,
            idx=index,
            column_idx=column_idx,
        )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 19:00:00                  #
# ================================================== #

from pygpt_net.plugin.base.config import BaseConfig, BasePlugin


class Config(BaseConfig):
    def __init__(self, plugin: BasePlugin = None, *args, **kwargs):
        super(Config, self).__init__(plugin)
        self.plugin = plugin

    def from_defaults(self, plugin: BasePlugin = None):
        """
        Set default options for plugin

        :param plugin: plugin instance
        """
        keys = {
            "enabled": "bool",
            "name": "text",
            "instruction": "textarea",
            "params": "textarea",
            "cmd": "textarea",
        }
        items = [
            {
                "enabled": True,
                "name": "example_cmd",
                "instruction": "execute tutorial test command by replacing 'hello' and 'world' params with some funny "
                               "words",
                "params": "hello, world",
                "cmd": 'echo "Response from {hello} and {world} at {_time}"',
            },
        ]
        desc = "Add your custom commands here, use {placeholders} to receive params, you can also use predefined " \
               "placeholders: {_time}, {_date}, {_datetime}, {_file}, {_home) "
        tooltip = "See the documentation for more details about examples, usage and list of predefined placeholders"
        plugin.add_option(
            "cmds",
            type="dict",
            value=items,
            label="Your custom commands",
            description=desc,
            tooltip=tooltip,
            keys=keys,
        )


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.20 21:00:00                  #
# ================================================== #

import json
import re
from datetime import datetime

from PySide6.QtCore import Slot

from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.core.events import Event, KernelEvent
from pygpt_net.item.ctx import CtxItem

from .config import Config
from .worker import Worker


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "cmd_history"
        self.name = "Context history (calendar, inline)"
        self.type = ["cmd.inline"]
        self.description = "Provides real-time access to context history database"
        self.prefix = "History"
        self.input_text = None
        self.allowed_cmds = [
            "get_ctx_list_in_date_range",
            "get_ctx_content_by_id",
            "count_ctx_in_date",
            "get_day_note",
            "add_day_note",
            "update_day_note",
            "remove_day_note",
        ]
        self.order = 100
        self.use_locale = True
        self.worker = None
        self.config = Config(self)
        self.init_options()

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx

        if name in [
            Event.CMD_SYNTAX_INLINE,  # inline is allowed
            Event.CMD_SYNTAX,
        ]:
            self.cmd_syntax(data)

        # append current time to system prompt if time plugin is not enabled
        elif name == Event.SYSTEM_PROMPT:
            data['value'] = self.on_system_prompt(data['value'])

        elif name == Event.POST_PROMPT:
            data['value'] = self.on_post_prompt(
                data['value'],
                ctx,
            )

        elif name == Event.USER_SEND:
            self.on_user_send(data['value'])

        elif name in [
            Event.CMD_INLINE,  # inline is allowed
            Event.CMD_EXECUTE,
        ]:
            self.cmd(
                ctx,
                data['commands'],
            )

    def on_system_prompt(self, prompt: str) -> str:
        """
        Event: SYSTEM_PROMPT

        :param prompt: prompt
        :return: updated prompt
        """
        # current time
        if not self.window.controller.plugins.is_type_enabled("time"):
            prompt += "\nCurrent time is: " + datetime.now().strftime('%A, %Y-%m-%d %H:%M:%S')
        return prompt

    def on_post_prompt(self, prompt: str, ctx: CtxItem) -> str:
        """
        Event: POST_PROMPT

        :param prompt: prompt
        :param ctx: context
        :return: updated prompt
        """
        # disable if internal context (autonomous mode)
        if ctx.internal:
            return prompt

        # hashtags
        if self.get_option_value("use_tags") and self.input_text:
            extra = self.handle_tags(self.input_text)
            if extra:
                prompt += "\n" + self.get_option_value("prompt_tag_system").format(
                    context=extra,
                )
        self.input_text = None  # reset input text
        return prompt

    def handle_tags(self, text: str) -> str:
        """
        Handle ID tags in input text

        :param text: input text
        :return: summary of contexts in JSON format
        """
        summary = ""
        contexts = []
        hashtags = self.extract_tags(text)
        ids = []
        for tag in hashtags:
            if tag.isdigit():
                id = int(tag)
                ids.append(id)
        for id in ids:
            try:
                prompt = self.get_option_value("prompt_tag_summary").format(
                    id=id,
                    query=self.input_text,
                )
            except Exception as e:
                self.log("Incorrect prompt: " + str(e))
                return ""
            summary = self.get_summary(id, prompt)
            if summary:
                self.log("@using: " + str(id))
                contexts.append({id: summary})
        if contexts:
            summary = json.dumps(contexts)
        return summary

    def extract_tags(self, text: str) -> list:
        """
        Extract tags from text

        :param text: text
        :return: list of tags
        """
        return re.findall(r"@(\d+)", text)

    def on_user_send(self, text: str):
        """
        Event: USER_SEND

        :param text: text
        """
        self.input_text = text

    def cmd_syntax(self, data: dict):
        """
        Event: CMD_SYNTAX

        :param data: event data dict
        """
        for option in self.allowed_cmds:
            if self.has_cmd(option):
                data['cmd'].append(self.get_cmd(option))  # append command

    @Slot()
    def handle_updated(self):
        """Handle updated signal"""
        self.window.controller.calendar.setup()

    def cmd(self, ctx: CtxItem, cmds: list):
        """
        Events: CMD_EXECUTE

        :param ctx: CtxItem
        :param cmds: commands dict
        """
        is_cmd = False
        my_commands = []
        for item in cmds:
            if item["cmd"] in self.allowed_cmds and self.has_cmd(item["cmd"]):
                my_commands.append(item)
                is_cmd = True

        if not is_cmd:
            return

        # set state: busy
        self.cmd_prepare(ctx, my_commands)

        worker = Worker()
        worker.from_defaults(self)
        worker.cmds = my_commands
        worker.ctx = ctx

        worker.signals.updated.connect(self.handle_updated)

        if not self.is_async(ctx):
            worker.run()
            return
        worker.run_async()

    def get_day_note(self, year: int, month: int, day: int) -> str:
        """
        Get day note

        :param year: year
        :param month: month
        :param day: day
        :return: day note
        """
        return self.window.core.calendar.load_note(year, month, day)

    def add_day_note(self, year: int, month: int, day: int, note: str) -> bool:
        """
        Add day note

        :param year: year
        :param month: month
        :param day: day
        :param note: note
        :return: True if success
        """
        return self.window.core.calendar.append_to_note(year, month, day, note)

    def update_day_note(self, year: int, month: int, day: int, note: str) -> bool:
        """
        Update day note

        :param year: year
        :param month: month
        :param day: day
        :param note: note
        :return: True if success
        """
        return self.window.core.calendar.update_note(year, month, day, note)

    def remove_day_note(self, year: int, month: int, day: int) -> bool:
        """
        Remove day note

        :param year: year
        :param month: month
        :param day: day
        :return: True if success
        """
        return self.window.core.calendar.remove_note(year, month, day)

    def get_list(self, range: str) -> list:
        """
        Get context list in date range

        :param range: date range
        :return: list of context
        """
        limit = int(self.get_option_value("ctx_items_limit"))
        return self.window.core.ctx.get_list_in_date_range(range, limit=limit)

    def get_summary(self, id: int, prompt: str) -> str:
        """
        Get context content summary by ID

        :param id: context ID
        :param prompt: prompt for summarization
        :return: context content summary
        """
        chunk_size = int(self.get_option_value("chunk_size"))
        items = self.window.core.ctx.get_items_by_id(id)
        if items is None or len(items) == 0:
            return ""
        content = "\n".join(items)
        chunks = self.to_chunks(content, chunk_size)  # it returns list of chunks
        summary = self.get_summarized_text(chunks, prompt)
        return summary

    def count_ctx_in_date(
            self,
            year: int = None,
            month: int = None,
            day: int = None
    ) -> dict:
        """
        Get context counters

        :param year: year
        :param month: month
        :param day: day
        :return: counters dict
        """
        return self.window.core.ctx.provider.get_ctx_count_by_day(
            year=year,
            month=month,
            day=day,
        )

    def to_chunks(
            self,
            text: str,
            chunk_size: int
    ) -> list:
        """
        Split text into chunks

        :param text: text to split
        :param chunk_size: chunk size
        :return: list of chunks
        """
        if text is None or text == "":
            return []
        return [text[i:i + chunk_size] for i in range(0, len(text), chunk_size)]

    def get_summarized_text(
            self,
            chunks: list,
            sys_prompt: str = None
    ) -> str:
        """
        Get summarized text from chunks, TODO: create external method for chunk summarization

        :param chunks: chunks of text
        :param sys_prompt: system prompt
        :return: summarized text
        """
        summary = []

        # get custom prompt if set
        max_tokens = int(self.get_option_value("summary_max_tokens"))

        # get model
        model = self.window.core.models.from_defaults()
        tmp_model = self.get_option_value("summary_model")
        if self.window.core.models.has(tmp_model):
            model = self.window.core.models.get(tmp_model)

        # summarize per chunk
        for chunk in chunks:
            try:
                bridge_context = BridgeContext(
                    prompt=chunk,
                    system_prompt=sys_prompt,
                    max_tokens=max_tokens,
                    model=model,
                    temperature=0.0,
                )
                event = KernelEvent(KernelEvent.CALL, {
                    'context': bridge_context,
                    'extra': {},
                })
                self.window.dispatch(event)
                response = event.data.get('response')
                if response is not None and response != "":
                    summary.append(response)
            except Exception as e:
                self.error(e)

        return "".join(summary)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict, Any, List

from pygpt_net.ui.widget.option.dictionary import OptionDict
from pygpt_net.utils import trans


class Dictionary:
    def __init__(self, window=None):
        """
        Dictionary field handler

        :param window: Window instance
        """
        self.window = window

    def apply(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ):
        """
        Apply values to dictionary

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        """
        if option["value"] is None:
            option["value"] = []
        values = list(option["value"])
        self.window.ui.config[parent_id][key].items = values  # replace model data list
        self.window.ui.config[parent_id][key].model.updateData(values)  # update model data

    def apply_row(
            self,
            parent_id: str,
            key: str,
            values: Dict[str, Any],
            idx: int
    ):
        """
        Apply data values to dictionary

        :param parent_id: Options parent ID
        :param key: Option key
        :param values: dictionary data values
        :param idx: row index
        """
        # if cmd type, remove .params suffix
        if key.endswith(".params"):
            key = key.replace(".params", "")
        self.window.ui.config[parent_id][key].update_item(idx, values)

    def get_value(
            self,
            parent_id: str,
            key: str,
            option: Dict[str, Any]
    ) -> List[Dict[str, Any]]:
        """
        Get dictionary values

        :param parent_id: Options parent ID
        :param key: Option key
        :param option: Option data
        :return: list with dictionary values
        """
        return self.window.ui.config[parent_id][key].model.items

    def delete_item(
            self,
            parent_object: OptionDict,
            id: str,
            force: bool = False,
            hooks: bool = True
    ):
        """
        Show delete item (from dict list) confirmation dialog or executes delete

        :param parent_object: parent object
        :param id: item id
        :param force: force delete
        :param hooks: run hooks
        """
        if not force:
            self.window.ui.dialogs.confirm(
                type="settings.dict.delete",
                id=id,
                msg=trans("settings.dict.delete.confirm"),
                parent_object=parent_object,
            )
            return

        # delete item
        if parent_object is not None:
            parent_object.delete_item_execute(id)

            # on update hooks
            if hooks:
                hook_name = "update.{}.{}".format(parent_object, id)
                if self.window.ui.has_hook(hook_name):
                    hook = self.window.ui.get_hook(hook_name)
                    try:
                        hook(id, {}, "dictionary")
                    except Exception as e:
                        self.window.core.debug.log(e)

    def to_options(
            self,
            parent_id: str,
            option: Dict[str, Any]
    ) -> Dict[str, Any]:
        """
        Convert dictionary items option to options

        :param parent_id: Parent ID
        :param option: dictionary items option
        :return: options dict
        """
        if "keys" not in option:
            return {}
        options = {}
        for key in option["keys"]:
            item = option["keys"][key]
            if isinstance(item, str):
                item = {
                    "label": parent_id + '.' + key,
                    "type": item,  # field type is provided as value in this case
                }
            options[key] = item
            if "label" not in options[key]:
                options[key]["label"] = key
                options[key]["label"] = parent_id + "." + options[key]["label"]
        return options

    def append_editor(
            self,
            id: str,
            option: Dict[str, Any],
            data: Dict[str, Any]
    ):
        """
        Apply dict item option sub-values to dict editor

        :param id: Owner object ID
        :param option: Option dict
        :param data: Option data
        """
        parent_id = "dictionary." + id
        sub_options = self.to_options(id, option)
        for key in sub_options:
            sub_option = sub_options[key]
            if key in data:
                sub_option["value"] = data[key]
            else:
                sub_option["value"] = ""
            self.window.controller.config.apply(
                parent_id=parent_id,
                key=key,
                option=sub_option,
            )

    def save_editor(
            self,
            option_key: str,
            parent: str,
            fields: Dict[str, Dict[str, Any]],
            hooks: bool = True
    ):
        """
        Save dict editor (called from dict editor dialog save button)

        :param option_key: Option key
        :param parent: Parent ID
        :param fields: Fields dict
        :param hooks: run hooks
        """
        values = {}
        dict_id = parent + "." + option_key
        dialog_id = "dictionary." + parent + "." + option_key  # dictionary parent ID
        idx = self.window.ui.dialog["editor." + dialog_id].idx  # editing record idx is stored in dialog idx
        for key in fields:
            value = self.window.controller.config.get_value(
                parent_id="dictionary." + dict_id,
                key=key,
                option=fields[key],
            )
            values[key] = value

        # update values in dictionary item on list in parent
        self.apply_row(
            parent,
            option_key,
            values,
            idx,
        )

        # close dialog
        self.window.ui.dialog['editor.' + dialog_id].close()

        # on update hooks
        if hooks:
            hook_name = "update.{}.{}".format(parent, option_key)
            if self.window.ui.has_hook(hook_name):
                hook = self.window.ui.get_hook(hook_name)
                try:
                    hook(option_key, [values], "dictionary")  # value must be list here
                except Exception as e:
                    self.window.core.debug.log(e)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 00:00:00                  #
# ================================================== #

import copy
from typing import Optional

from PySide6.QtGui import QAction, QIcon
from packaging.version import Version

from pygpt_net.item.prompt import PromptItem
from pygpt_net.provider.core.prompt.json_file import JsonFileProvider
from pygpt_net.utils import trans


class Custom:
    def __init__(self, window=None):
        """
        Custom prompts core

        :param window: Window instance
        """
        self.window = window
        self.provider = JsonFileProvider(window)
        self.items = {}
        self.current = None
        self.initiated = False

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version) -> bool:
        """
        Patch provider data

        :param app_version: app version
        :return: True if data was patched
        """
        return self.provider.patch(app_version)

    def select(self, id: str):
        """
        Select prompt by uuid

        :param id: id
        """
        if id in self.items:
            self.current = id

    def count(self) -> int:
        """
        Count prompts

        :return: prompts count
        """
        return len(self.items)

    def get_ids(self) -> list:
        """
        Get items IDs

        :return: items UUIDs
        """
        return list(self.items.keys())

    def get_id_by_idx(self, idx: int) -> Optional[str]:
        """
        Get ID by index

        :param idx: index
        :return: file ID
        """
        i = 0
        for id in self.get_ids():
            if i == idx:
                return id
            i += 1

    def get_by_id(self, id: str) -> Optional[PromptItem]:
        """
        Return prompt by ID

        :param id: file id
        :return: PromptItem
        """
        if id in self.items:
            return self.items[id]

    def get_by_idx(self, idx: int) -> Optional[PromptItem]:
        """
        Return item by index

        :param idx: item index
        :return: PromptItem or None
        """
        id = self.get_id_by_idx(idx)
        if id is not None:
            return self.items[id]

    def get_all(self) -> dict:
        """
        Return all items

        :return: prompts items dict
        """
        return self.items

    def new(
            self,
            name: Optional[str] = None,
            content: Optional[str] = None,
            auto_save: bool = True
    ) -> PromptItem:
        """
        Create new prompt

        :param name: name
        :param content: content
        :param auto_save: auto_save
        :return: PromptItem
        """
        prompt = self.create()
        prompt.name = name
        prompt.content = content

        self.items[prompt.id] = prompt
        self.current = prompt.id

        if auto_save:
            self.save()

        return prompt

    def build(self) -> PromptItem:
        """
        Build prompt

        :return: PromptItem
        """
        prompt = PromptItem()
        prompt.name = None
        prompt.content = None
        return prompt

    def create(self) -> PromptItem:
        """
        Create prompt item

        :return: PromptItem
        """
        prompt = self.build()
        id = self.provider.create(prompt)
        prompt.id = id
        return prompt

    def add(self, item: PromptItem):
        """
        Add item to prompts

        :param item: item to add to prompts
        """
        id = item.id
        self.items[id] = item  # add item to prompts

        # save prompts
        self.save()

    def has(self) -> bool:
        """
        Check id mode has prompts

        :return: True if exists
        """
        return len(self.items) > 0

    def delete(self, id: str):
        """
        Delete prompt by file_id

        :param id: file id
        """
        if id in self.items:
            del self.items[id]
            self.save()

    def delete_all(self):
        """
        Delete all prompts
        """
        self.clear()
        self.provider.truncate()

    def clear(self):
        """
        Clear all prompts

        :param mode: mode
        :param remove_local: remove local copy
        """
        self.items = {}

    def clear_all(self):
        """
        Clear all prompts
        """
        self.items = {}

    def rename_file(self, id: str, name: str):
        """
        Update name

        :param id: file id
        :param name: new name
        """
        data = self.get_by_id(id)
        data.name = name
        self.save()

    def make_json_list(self, prompts: dict) -> dict:
        """
        Make json list

        :param prompts: prompts
        :return: json list
        """
        result = {}
        for id in prompts:
            prompt = prompts[id]
            result[id] = {
                'name': prompt.name,
                'content': prompt.content,
            }
        return result

    def reload(self):
        """Reload prompts"""
        self.items = self.provider.load()

    def load(self):
        """Load prompts"""
        self.items = self.provider.load()

    def save(self):
        """Save prompts"""
        data = copy.deepcopy(self.items)  # copy to avoid changing original data
        self.provider.save(data)

    def init(self):
        """Init prompts"""
        if not self.initiated:
            self.load()
            self.initiated = True

    def to_menu_options(self, menu, parent: str = "global"):
        """
        Convert prompts to menu options

        :param menu: menu
        :param parent: parent
        """
        self.init()

        # at first clear menu
        menu.addSeparator()
        submenu = menu.addMenu(trans("preset.prompt.paste_custom"))
        submenu.clear()
        letter_submenus = {}

        # add submenus for each letter
        for id in self.items.keys():
            value = self.items[id]
            if value.name is None or value.name == "":
                continue
            letter = value.name[0].upper()
            if letter not in letter_submenus:
                letter_submenus[letter] = submenu.addMenu(letter)

        # add prompts to letter submenus
        for id in self.items.keys():
            value = self.items[id]
            if value.name is None or value.name == "":
                continue
            letter = value.name[0].upper()
            item_submenu = letter_submenus[letter].addMenu(value.name)

            action_use = QAction(QIcon(":/icons/paste.svg"), trans("preset.prompt.use"), menu)
            action_use.triggered.connect(lambda checked=False, key=id: self.window.controller.presets.paste_custom_prompt(key, parent))
            action_use.setToolTip(value.content)

            action_rename = QAction(QIcon(":/icons/edit.svg"), trans("preset.prompt.rename"), menu)
            action_rename.triggered.connect(lambda checked=False, key=id: self.window.controller.presets.rename_prompt(key))

            action_delete = QAction(QIcon(":/icons/delete.svg"), trans("preset.prompt.delete"), menu)
            action_delete.triggered.connect(
                lambda checked=False, key=id: self.window.controller.presets.delete_prompt(key))

            item_submenu.addAction(action_use)
            item_submenu.addAction(action_rename)
            item_submenu.addAction(action_delete)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 21:00:00                  #
# ================================================== #

from PySide6.QtCore import Slot

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = BaseSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None

    @Slot()
    def run(self):
        responses = []
        msg = None
        for item in self.cmds:
            if self.is_stopped():
                break
            response = None
            try:
                if item["cmd"] == "get_context":
                    response = self.cmd_get_context(item)

                if response:
                    responses.append(response)

            except Exception as e:
                msg = "Error: {}".format(e)
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

        # update status
        if msg is not None:
            self.status(msg)

    def cmd_get_context(self, item: dict) -> dict:
        """
        Get context for given query

        :param item: command item
        :return: response item
        """
        question = self.get_param(item, "query")
        self.status("Please wait... querying: {}...".format(question))
        # at first, try to get from retrieval
        response = self.plugin.get_from_retrieval(question)
        if response is not None and response != "":
            self.log("Found using retrieval...")
            context = "ADDITIONAL CONTEXT (response from DB):\n--------------------------------\n" + response
            extra = {
                "context": context,
            }
            return self.make_response(item, response, extra=extra)

        content, doc_ids, metas = self.plugin.query(question)  # send question to Llama-index
        result = content
        context = "ADDITIONAL CONTEXT (response from DB):\n--------------------------------\n" + content
        if doc_ids:
            self.ctx.doc_ids = doc_ids  # store doc_ids in context

        extra = {
            "doc_ids": doc_ids,
            "metas": metas,
            "context": context,
        }
        return self.make_response(item, result, extra=extra)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.20 21:00:00                  #
# ================================================== #

from pygpt_net.core.events import KernelEvent
from pygpt_net.core.bridge.context import BridgeContext
from pygpt_net.item.ctx import CtxItem


class Summarizer:
    def __init__(self, window=None):
        """
        Summarizer

        :param window: Window instance
        """
        self.window = window

    def summary_ctx(self, ctx: CtxItem) -> str:
        """
        Summarize conversation begin

        :param ctx: context item (CtxItem)
        :return: response text (generated summary)
        """
        model = self.window.core.models.from_defaults()
        system_prompt = self.window.core.prompt.get('ctx.auto_summary.system')
        text = (self.window.core.prompt.get('ctx.auto_summary.user').
                replace("{input}", str(ctx.input)).
                replace("{output}", str(ctx.output)))

        # custom model for auto summary
        if self.window.core.config.get('ctx.auto_summary.model') is not None \
                and self.window.core.config.get('ctx.auto_summary.model') != "":
            tmp_model = self.window.core.config.get('ctx.auto_summary.model')
            if self.window.core.models.has(tmp_model):
                model = self.window.core.models.get(tmp_model)

        # quick call OpenAI API
        bridge_context = BridgeContext(
            ctx=ctx,
            prompt=text,
            system_prompt=system_prompt,
            model=model,
            max_tokens=500,
            temperature=0.0,
        )
        event = KernelEvent(KernelEvent.CALL, {
            'context': bridge_context,
            'extra': {},
            'response': None,
        })
        self.window.dispatch(event)
        response = event.data.get('response')
        if response is not None:
            return response



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.18 21:00:00                  #
# ================================================== #

import serial
import time

from PySide6.QtCore import Slot

from pygpt_net.plugin.base.worker import BaseWorker, BaseSignals


class WorkerSignals(BaseSignals):
    pass  # add custom signals here


class Worker(BaseWorker):
    def __init__(self, *args, **kwargs):
        super(Worker, self).__init__()
        self.signals = WorkerSignals()
        self.args = args
        self.kwargs = kwargs
        self.plugin = None
        self.cmds = None
        self.ctx = None
        self.msg = None

    @Slot()
    def run(self):
        responses = []
        for item in self.cmds:
            if self.is_stopped():
                break
            response = None
            try:
                if item["cmd"] in self.plugin.allowed_cmds and self.plugin.has_cmd(item["cmd"]):

                    # serial: send text command
                    if item["cmd"] == "serial_send":
                        response = self.cmd_serial_send(item)

                    # serial: send raw bytes command
                    elif item["cmd"] == "serial_send_bytes":
                        response = self.cmd_serial_send_bytes(item)

                    # serial: read data from USB port
                    elif item["cmd"] == "serial_read":
                        response = self.cmd_serial_read(item)

                    if response:
                        responses.append(response)

            except Exception as e:
                responses.append(
                    self.make_response(
                        item,
                        self.throw_error(e)
                    )
                )

        # send response
        if len(responses) > 0:
            self.reply_more(responses)

        if self.msg is not None:
            self.status(self.msg)

    def cmd_serial_send(self, item: dict) -> dict:
        """
        Send command to USB port

        :param item: command item
        :return: response item
        """
        port = self.plugin.get_option_value("serial_port")
        speed = self.plugin.get_option_value("serial_bps")
        timeout = self.plugin.get_option_value("timeout")
        sleep = self.plugin.get_option_value("sleep")
        self.log("Using serial port: {} @ {} bps".format(port, speed))
        try:
            self.msg = "Sending command to USB port: {}".format(
                self.get_param(item, "command"),
            )
            self.log(self.msg)
            result = self.send_command(
                port,
                speed,
                self.get_param(item, "command"),
                timeout=timeout,
                sleep=sleep,
            )
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_serial_send_bytes(self, item: dict) -> dict:
        """
        Send raw bytes to USB port

        :param item: command item
        :return: response item
        """
        port = self.plugin.get_option_value("serial_port")
        speed = self.plugin.get_option_value("serial_bps")
        timeout = self.plugin.get_option_value("timeout")
        sleep = self.plugin.get_option_value("sleep")
        self.log("Using serial port: {} @ {} bps".format(port, speed))
        try:
            self.msg = "Sending binary data to USB port: {}".format(
                self.get_param(item, "bytes"),
            )
            self.log(self.msg)
            result = self.send_binary_data(
                port,
                speed,
                int(self.get_param(item, "bytes")),
                timeout=timeout,
                sleep=sleep,
            )
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def cmd_serial_read(self, item: dict) -> dict:
        """
        Read data from USB port

        :param item: command item
        :return: response item
        """
        port = self.plugin.get_option_value("serial_port")
        speed = self.plugin.get_option_value("serial_bps")
        timeout = self.plugin.get_option_value("timeout")
        duration = int(item["params"]['duration']) \
            if self.has_param(item, "duration") else 3
        self.log("Using serial port: {} @ {} bps".format(port, speed))
        try:
            self.msg = "Reading data from USB port..."
            self.log(self.msg)
            result = self.read_data(
                port,
                speed,
                timeout=timeout,
                duration=duration,
            )
            self.log("Response: {}".format(result))
        except Exception as e:
            result = self.throw_error(e)
        return self.make_response(item, result)

    def send_command(
            self,
            port: str,
            speed: int,
            command: str,
            timeout: int = 1,
            sleep: int = 2
    ) -> str:
        """
        Send command to USB port

        :param port: USB port name, e.g. /dev/ttyACM0
        :param speed: Port connection speed, in bps, default: 9600
        :param command: Command to send
        :param timeout: Timeout in seconds
        :param sleep: Sleep time in seconds
        :return: Response from USB port
        """
        ser = serial.Serial(port, speed, timeout=timeout)
        time.sleep(sleep)
        ser.write((command + '\n').encode())
        ser.flush()
        return ser.readline().decode().strip()

    def send_binary_data(
            self,
            port: str,
            speed: int,
            data: int,
            timeout: int = 1,
            sleep: int = 2
    ) -> str:
        """
        Send command to USB port

        :param port: USB port name, e.g. /dev/ttyACM0
        :param speed: Port connection speed, in bps, default: 9600
        :param data: Data to send
        :param timeout: Timeout in seconds
        :param sleep: Sleep time in seconds
        :return: Response from USB port
        """
        ser = serial.Serial(port, speed, timeout=timeout)
        time.sleep(sleep)
        ser.write(bytes(data))
        ser.flush()
        return ser.readline().decode().strip()

    def read_data(
            self,
            port: str,
            speed: int,
            timeout: int = 1,
            duration: int = 3
    ) -> str:
        """
        Read data from USB port

        :param port: USB port name, e.g. /dev/ttyACM0
        :param speed: Port connection speed, in bps, default: 9600
        :param timeout: Timeout in seconds
        :param duration: Duration in seconds
        :return: Response from USB port
        """
        data = ""
        ser = serial.Serial(port, speed, timeout=timeout)
        end_time = time.time() + duration
        while time.time() < end_time:
            if ser.in_waiting > 0:
                data += ser.readline().decode().strip()
        return data

    def prepare_request(self, item) -> dict:
        """
        Prepare request item for result

        :param item: item with parameters
        :return: request item
        """
        return {"cmd": item["cmd"]}



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.03.02 19:00:00                  #
# ================================================== #

import copy
import uuid
from typing import Optional, Tuple, Dict

from packaging.version import Version
from pygpt_net.core.types import (
    MODE_AGENT,
    MODE_AGENT_LLAMA,
    MODE_ASSISTANT,
    MODE_AUDIO,
    MODE_CHAT,
    MODE_COMPLETION,
    MODE_EXPERT,
    MODE_IMAGE,
    MODE_LANGCHAIN,
    MODE_LLAMA_INDEX,
    MODE_VISION,
    MODE_RESEARCH,
)
from pygpt_net.item.preset import PresetItem
from pygpt_net.provider.core.preset.json_file import JsonFileProvider


class Presets:

    def __init__(self, window=None):
        """
        Presets core

        :param window: Window instance
        """
        self.window = window
        self.provider = JsonFileProvider(window)
        self.items = {}

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version):
        """
        Patch provider data

        :param app_version: app version
        """
        self.provider.patch(app_version)

    def build(self) -> PresetItem:
        """
        Build empty preset

        :return: empty preset
        """
        preset = PresetItem()
        preset.uuid = str(uuid.uuid4())
        return preset

    def append_current(self):
        """Append current presets"""
        # create empty presets
        curr_chat = self.build()
        curr_completion = self.build()
        curr_img = self.build()
        curr_vision = self.build()
        curr_langchain = self.build()
        curr_assistant = self.build()
        curr_llama = self.build()
        curr_agent = self.build()
        curr_agent_llama = self.build()
        curr_expert = self.build()
        curr_audio = self.build()

        # prepare ids
        id_chat = 'current.chat'
        id_completion = 'current.completion'
        id_img = 'current.img'
        id_vision = 'current.vision'
        id_langchain = 'current.langchain'
        id_assistant = 'current.assistant'
        id_llama = 'current.llama_index'
        id_agent = 'current.agent'
        id_agent_llama = 'current.agent_llama'
        id_expert = 'current.expert'
        id_audio = 'current.audio'

        # set default initial prompt for chat mode
        curr_chat.prompt = self.window.core.prompt.get('default')

        # get data from presets if exists
        if id_chat in self.items:
            curr_chat = self.items[id_chat]
        if id_completion in self.items:
            curr_completion = self.items[id_completion]
        if id_img in self.items:
            curr_img = self.items[id_img]
        if id_vision in self.items:
            curr_vision = self.items[id_vision]
        if id_langchain in self.items:
            curr_langchain = self.items[id_langchain]
        if id_assistant in self.items:
            curr_assistant = self.items[id_assistant]
        if id_llama in self.items:
            curr_llama = self.items[id_llama]
        if id_agent in self.items:
            curr_agent = self.items[id_agent]
        if id_agent_llama in self.items:
            curr_agent_llama = self.items[id_agent_llama]
        if id_expert in self.items:
            curr_expert = self.items[id_expert]
        if id_audio in self.items:
            curr_audio = self.items[id_audio]

        # allow usage in specific mode
        curr_chat.chat = True
        curr_completion.completion = True
        curr_img.img = True
        curr_vision.vision = True
        curr_langchain.langchain = True
        curr_assistant.assistant = True
        curr_llama.llama_index = True
        curr_agent.agent = True
        curr_agent_llama.agent_llama = True
        curr_expert.expert = True
        curr_audio.audio = True

        # always apply default name
        curr_chat.name = '*'
        curr_completion.name = '*'
        curr_img.name = '*'
        curr_vision.name = '*'
        curr_langchain.name = '*'
        curr_assistant.name = '*'
        curr_llama.name = '*'
        curr_agent.name = '*'
        curr_agent_llama.name = '*'
        curr_expert.name = '*'
        curr_audio.name = '*'

        # append at first position
        self.items = {
            id_chat: curr_chat,
            id_completion: curr_completion,
            id_img: curr_img,
            id_vision: curr_vision,
            id_langchain: curr_langchain,
            id_assistant: curr_assistant,
            id_llama: curr_llama,
            id_agent: curr_agent,
            id_agent_llama: curr_agent_llama,
            id_expert: curr_expert,
            id_audio: curr_audio,
            **self.items
        }

    def exists(self, id: str) -> bool:
        """
        Check if preset exists

        :param id: preset id
        :return: True if exists
        """
        if id in self.items:
            return True
        return False

    def exists_uuid(self, uuid: str) -> bool:
        """
        Check if preset exists

        :param uuid: preset uuid
        :return: True if exists
        """
        for id in self.items:
            if self.items[id].uuid == uuid:
                return True
        return False

    def enable(self, id: str):
        """
        Enable preset

        :param id: preset id
        """
        if id in self.items:
            self.items[id].enabled = True

    def disable(self, id: str):
        """
        Disable preset

        :param id: preset id
        """
        if id in self.items:
            self.items[id].enabled = False

    def get_first_mode(self, id: str) -> Optional[str]:
        """
        Return first mode for preset

        :param id: preset id
        :return: mode name
        """
        preset = self.items[id]
        if preset.chat:
            return MODE_CHAT
        if preset.completion:
            return MODE_COMPLETION
        if preset.img:
            return MODE_IMAGE
        if preset.vision:
            return MODE_VISION
        if preset.langchain:
            return MODE_LANGCHAIN
        if preset.assistant:
            return MODE_ASSISTANT
        if preset.llama_index:
            return MODE_LLAMA_INDEX
        if preset.agent:
            return MODE_AGENT
        if preset.agent_llama:
            return MODE_AGENT_LLAMA
        if preset.expert:
            return MODE_EXPERT
        if preset.audio:
            return MODE_AUDIO
        if preset.research:
            return MODE_RESEARCH
        return None

    def has(self, mode: str, id: str) -> bool:
        """
        Check if preset for mode exists

        :param mode: mode name
        :param id : preset id
        :return: True if exists
        """
        presets = self.get_by_mode(mode)
        if id in presets:
            return True
        return False

    def get_by_idx(self, idx: int, mode: str) -> str:
        """
        Return preset by index

        :param idx: index
        :param mode: mode
        :return: preset id
        """
        presets = self.get_by_mode(mode)
        return list(presets.keys())[idx]

    def get_by_id(self, mode: str, id: str) -> Optional[PresetItem]:
        """
        Return preset by id

        :param mode: mode name
        :param id: preset id
        :return: preset item
        """
        presets = self.get_by_mode(mode)
        if id in presets:
            return presets[id]
        return None

    def get_by_uuid(self, uuid: str) -> Optional[PresetItem]:
        """
        Return preset by UUID

        :param uuid: preset UUID
        :return: preset item
        """
        for id in self.items:
            if self.items[id].uuid == uuid:
                return self.items[id]
        return None

    def get_by_mode(self, mode: str) -> Dict[str, PresetItem]:
        """
        Return presets for mode

        :param mode: mode name
        :return: presets dict for mode
        """
        presets = {}
        for id in self.items:
            if (mode == MODE_CHAT and self.items[id].chat) \
                    or (mode == MODE_COMPLETION and self.items[id].completion) \
                    or (mode == MODE_IMAGE and self.items[id].img) \
                    or (mode == MODE_VISION and self.items[id].vision) \
                    or (mode == MODE_LANGCHAIN and self.items[id].langchain) \
                    or (mode == MODE_ASSISTANT and self.items[id].assistant) \
                    or (mode == MODE_LLAMA_INDEX and self.items[id].llama_index) \
                    or (mode == MODE_AGENT and self.items[id].agent) \
                    or (mode == MODE_AGENT_LLAMA and self.items[id].agent_llama) \
                    or (mode == MODE_EXPERT and self.items[id].expert) \
                    or (mode == MODE_RESEARCH and self.items[id].research) \
                    or (mode == MODE_AUDIO and self.items[id].audio):
                presets[id] = self.items[id]
        return presets

    def get_idx_by_id(self, mode: str, id: str) -> int:
        """
        Return preset index

        :param mode: mode name
        :param id: preset id
        :return: preset idx
        """
        presets = self.get_by_mode(mode)
        i = 0
        for key in presets:
            if key == id:
                return i
            i += 1
        return 0

    def get_default(self, mode: str) -> Optional[str]:
        """
        Return default preset for mode

        :param mode: mode name
        :return: default prompt name
        """
        presets = self.get_by_mode(mode)
        if len(presets) == 0:
            return None
        return list(presets.keys())[0]

    def get_duplicate_name(self, id: str) -> Tuple[str, str]:
        """
        Prepare name for duplicated preset

        :param id: preset id
        :return: new ID, new name
        """
        old_name = self.items[id].name
        new_id = id
        new_name = old_name
        while True:
            new_id = new_id + '_copy'
            new_name = new_name + ' copy'
            if new_id not in self.items:
                return new_id, new_name

    def duplicate(self, id: str) -> str:
        """
        Make preset duplicate

        :param id: preset id
        :return: duplicated preset ID
        """
        prev_id = id
        id, name = self.get_duplicate_name(id)
        self.items[id] = copy.deepcopy(self.items[prev_id])
        self.items[id].name = name
        self.items[id].filename = id
        self.items[id].uuid = str(uuid.uuid4())  # generate new uuid
        self.sort_by_name()
        return id

    def remove(
            self,
            id: str,
            remove_file: bool = True
    ):
        """
        Delete preset

        :param id: preset id
        :param remove_file: also remove preset JSON config file
        """
        if id in self.items:
            self.items.pop(id)

        if remove_file:
            self.provider.remove(id)
            # self.load_presets()

    def sort_by_name(self):
        """Sort presets by name"""
        self.items = dict(
            sorted(
                self.items.items(),
                key=lambda item: item[1].name
            )
        )

    def update(self, preset: PresetItem):
        """
        Update preset

        :param preset: preset item
        """
        self.items[preset.filename] = preset

    def update_and_save(self, preset: PresetItem):
        """
        Update preset

        :param preset: preset item
        """
        self.items[preset.filename] = preset
        self.save(preset.filename)

    def get_all(self) -> Dict[str, PresetItem]:
        """
        Return all presets

        :return: presets dict
        """
        return self.items

    def restore(self, mode: str):
        """
        Restore default preset for mode

        :param mode: mode name
        """
        base = self.load_base()
        id = "current." + mode
        if id in base:
            self.items[id] = base[id]
            self.save(id)

    def load_base(self) -> Dict[str, PresetItem]:
        """
        Load base presets

        :return: base presets
        """
        return self.provider.load_base()

    def load(self):
        """Load presets templates"""
        self.items = self.provider.load()
        self.patch_empty()  # patch empty UUIDs and filenames
        self.patch_duplicated()  # patch duplicated UUIDs

        # sort presets
        self.sort_by_name()
        self.append_current()

    def save(self, id: str):
        """
        Save preset

        :param id: preset id
        """
        if id not in self.items:
            return
        self.provider.save(id, self.items[id])

    def save_all(self):
        """Save all presets"""
        self.provider.save_all(self.items)

    def add_expert(
            self,
            agent_uuid: str,
            expert_uuid: str
    ):
        """
        Add expert to agent

        :param agent_uuid: agent uuid
        :param expert_uuid: expert uuid
        """
        agent = self.get_by_uuid(agent_uuid)
        if agent is None:
            return
        if expert_uuid not in agent.experts:
            agent.experts.append(expert_uuid)
        self.save(agent.filename)

    def remove_expert(
            self,
            agent_uuid: str,
            expert_uuid: str
    ):
        """
        Remove expert from agent
        :param agent_uuid: agent uuid
        :param expert_uuid: expert uuid
        """
        agent = self.get_by_uuid(agent_uuid)
        if agent is None:
            return
        if expert_uuid in agent.experts:
            agent.experts.remove(expert_uuid)
        self.save(agent.filename)

    def patch_empty(self):
        """Patch UUIDs for all presets"""
        patched = False
        for id in self.items:
            if self.items[id].uuid is None:
                self.items[id].uuid = str(uuid.uuid4())
                patched = True
            if self.items[id].filename is None or self.items[id].filename == "":
                self.items[id].filename = id
                patched = True
        if patched:
            self.save_all()

    def patch_duplicated(self):
        """Patch UUIDs for all presets"""
        patched = False
        uuids = []
        for id in self.items:
            if self.items[id].uuid in uuids:
                self.items[id].uuid = str(uuid.uuid4())
                patched = True
            uuids.append(self.items[id].uuid)
        if patched:
            self.save_all()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.26 02:00:00                  #
# ================================================== #

class AttachmentsDebug:
    def __init__(self, window=None):
        """
        Attachments debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'attachments'

    def update(self):
        """Update debug window."""
        self.window.core.debug.begin(self.id)

        for mode in self.window.core.modes.all:
            self.window.core.debug.add(self.id, '[' + mode + ']', '')
            attachments = self.window.core.attachments.get_all(mode)
            for key in list(attachments):
                attachment = attachments[key]
                data = {
                    'path': attachment.path,
                    'id': attachment.id,
                    'name': attachment.name,
                    'remote': attachment.remote,
                    'send': attachment.send,
                    'key': key,
                    'mode': mode,
                    'type': attachment.type,
                    'consumed': attachment.consumed,
                    'extra': attachment.extra,
                }
                self.window.core.debug.add(self.id, attachment.name, str(data))

        self.window.core.debug.end(self.id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.25 02:00:00                  #
# ================================================== #

import os
import json
import re
import docker
import io
import tarfile

from jupyter_client import BlockingKernelClient
from docker.errors import DockerException

class DockerKernel:
    def __init__(self, plugin = None):
        self.plugin = plugin
        self.kernel_file = ".interpreter.kernel.json"
        self.client = None
        self.container_name = "pygpt_ipython_kernel_container"
        self.image_name = "pygpt_ipython_kernel"
        self.initialized = False
        self.key = "19749810-8febfa748186a01da2f7b28c"
        self.bind_address = "0.0.0.0"
        self.conn_address = "127.0.0.1"
        self.ports = {
            "shell": 5555,
            "iopub": 5556,
            "stdin": 5557,
            "control": 5558,
            "hb": 5559,
        }
        self.signals = None

    def get_dockerfile(self) -> str:
        """
        Get the Dockerfile for the IPython kernel container.

        :return: Dockerfile.
        """
        return self.plugin.get_option_value('ipython_dockerfile')

    def get_key(self) -> str:
        """
        Get the key for the IPython kernel.

        :return: Key.
        """
        return self.plugin.get_option_value('ipython_session_key')

    def get_image_name(self) -> str:
        """
        Get the image name for the IPython kernel.

        :return: Image name.
        """
        return self.plugin.get_option_value('ipython_image_name')

    def get_container_name(self) -> str:
        """
        Get the container name for the IPython kernel.

        :return: Container name.
        """
        return self.plugin.get_option_value('ipython_container_name')

    def create_docker_context(self, dockerfile: str) -> io.BytesIO:
        """
        Create a Docker context with the specified Dockerfile content.

        :param dockerfile: Dockerfile content.
        :return: Docker context.
        """
        tar_stream = io.BytesIO()
        with tarfile.open(fileobj=tar_stream, mode='w') as tar:
            dockerfile_info = tarfile.TarInfo('Dockerfile')
            dockerfile_data = dockerfile.encode('utf-8')
            dockerfile_info.size = len(dockerfile_data)
            tar.addfile(dockerfile_info, io.BytesIO(dockerfile_data))
        tar_stream.seek(0)
        return tar_stream

    def is_image(self):
        """Check if the Docker image for the IPython kernel exists."""
        client = self.get_docker_client()
        try:
            client.images.get(self.get_image_name())
            return True
        except docker.errors.ImageNotFound:
            return False

    def restart(self):
        """Restart the container."""
        self.restart_container(self.get_container_name())

    def build_image(self):
        """Build the Docker image for the IPython kernel."""
        client = self.get_docker_client()
        context = self.create_docker_context(self.get_dockerfile())
        self.log("Please wait... Building the Docker image...")
        image, logs = client.images.build(
            fileobj=context,
            custom_context=True,
            rm=True,
            tag=self.get_image_name(),
        )
        for chunk in logs:
            if 'stream' in chunk:
                self.log(chunk['stream'].strip())

    def init(self, force: bool = False):
        """
        Initialize the IPython kernel client.

        :param force: Force reinitialization.
        """
        if self.initialized and not force:
            return
        self.prepare_local_data_dir()
        self.start_container(self.get_container_name())
        self.prepare_conn()
        self.client = BlockingKernelClient(connection_file=self.get_kernel_file_path())
        self.client.load_connection_file()
        self.client.start_channels()
        self.client.wait_for_ready()
        self.log("Connected to IPython kernel.")
        self.initialized = True
        self.log("IPython kernel is ready.")

    def prepare_local_data_dir(self):
        """
        Prepare the local data directory.
        """
        local_data_dir = self.get_local_data_dir()
        try:
            os.makedirs(local_data_dir)
        except FileExistsError:
            pass

    def get_docker_client(self) -> docker.DockerClient:
        """
        Get the Docker client.

        :return: Docker client.
        """
        return docker.from_env()

    def prepare_conn(self):
        """Prepare the connection file."""
        ports = self.get_ports()
        conn = {
            "shell_port": ports["shell"],
            "iopub_port": ports["iopub"],
            "stdin_port": ports["stdin"],
            "control_port": ports["control"],
            "hb_port": ports["hb"],
            "ip": self.get_conn_address(),
            "key": self.get_key(),
            "transport": "tcp",
            "signature_scheme": "hmac-sha256"
        }
        with open(self.get_kernel_file_path(), "w") as f:
            json.dump(conn, f)

    def process_message(self, msg: dict) -> str:
        """
        Process the message from the IPython kernel.

        :param msg: Message from the IPython kernel.
        :return: Processed message.
        """
        msg_type = msg['msg_type']
        content = msg['content']
        if msg_type == 'stream':
            # standard output and error
            return content['text']
        elif msg_type == 'display_data':
            # display data
            data = content['data']
            if 'text/plain' in data:
                return data['text/plain']
        elif msg_type == 'execute_result':
            # execution result
            data = content['data']
            if 'text/plain' in data:
                return data['text/plain']
        elif msg_type == 'error':
            # execution errors
            return "Error executing code:" + "\n".join(content['traceback'])
        return ""

    def end(self, all: bool = False):
        """
        Stop the IPython kernel.

        :param all: Stop the container as well.
        """
        self.client.stop_channels()  # stop the client
        if all:
            self.stop_container(self.get_container_name())

    def stop_container(self, name: str):
        """
        Stop the Docker container.

        :param name: Container name.
        """
        client = self.get_docker_client()
        try:
            container = client.containers.get(name)
            container.stop()
            container.remove()
        except docker.errors.NotFound:
            self.log(f"Container '{name}' not found.")

    def run_container(self, name: str) -> bool:
        """
        Run the Docker container.

        :param name: Container name.
        :return: True if the container was started successfully, False otherwise.
        """
        client = self.get_docker_client()
        ports = self.get_ports()
        # at first, check for image
        if not self.is_image():
            self.build_image()

        # run the container
        try:
            print("Running container {}...".format(name))
            self.prepare_conn()
            local_data_dir = self.get_local_data_dir()
            client.containers.run(
                self.get_image_name(),
                name=name,
                ports={
                    '5555/tcp': ports['shell'],
                    '5556/tcp': ports['iopub'],
                    '5557/tcp': ports['stdin'],
                    '5558/tcp': ports['control'],
                    '5559/tcp': ports['hb'],
                },
                # bind /data directory in container to the local data directory
                volumes={
                    local_data_dir: {
                        'bind': '/data',
                        'mode': 'rw',
                    }
                },
                detach=True,
            )
            return True
        except docker.errors.APIError as e:
            self.log(f"Error running container: {e}")
            return False

    def start_container(self, name: str):
        """
        Start the Docker container.

        :param name: Container name.
        """
        client = self.get_docker_client()
        try:
            client.containers.get(name)
        except docker.errors.NotFound:
            self.log(f"Container '{name}' not found. Creating new one...")
            self.log(f"Creating a new container: '{name}'...")
            self.run_container(name)
            self.log("Container has been started.")

    def restart_container(self, name: str):
        """
        Restart the Docker container.

        :param name: Container name.
        """
        client = self.get_docker_client()
        try:
            container = client.containers.get(name)
            self.log(f"Stopping container '{name}'...")
            container.stop()
            container.remove()
            self.log(f"Container '{name}' has been stopped and removed.")
        except docker.errors.NotFound:
            self.log(f"Container '{name}' not found. Nothing stopped.")

        self.log(f"Creating a new container: '{name}'...")
        self.run_container(name)
        self.log("Container has been started.")

    def get_conn_address(self) -> str:
        """
        Get the connection address.

        :return: Connection address.
        """
        return self.plugin.get_option_value('ipython_conn_addr')

    def get_bind_address(self) -> str:
        """
        Get the bind address.

        :return: Bind address.
        """
        return self.bind_address

    def get_ports(self) -> dict:
        """
        Get the ports.

        :return: Ports.
        """
        ports = {}
        ports['shell'] = self.plugin.get_option_value('ipython_port_shell')
        ports['iopub'] = self.plugin.get_option_value('ipython_port_iopub')
        ports['stdin'] = self.plugin.get_option_value('ipython_port_stdin')
        ports['control'] = self.plugin.get_option_value('ipython_port_control')
        ports['hb'] = self.plugin.get_option_value('ipython_port_hb')
        return ports

    def get_kernel_file_path(self) -> str:
        """
        Get the kernel file path.

        :return: Kernel file path.
        """
        return os.path.join(self.plugin.window.core.config.get_user_dir("data"), self.kernel_file)

    def get_local_data_dir(self) -> str:
        """
        Get the local data directory.

        :return: Local data directory.
        """
        return os.path.join(self.plugin.window.core.config.get_user_dir("data"), "ipython")

    def execute(self, code: str, current: bool = False) -> str:
        """
        Execute the code in the IPython kernel.

        :param code: Python code to execute.
        :param current: Use the current kernel.
        :return: Output from the kernel.
        """
        self.init()
        if not current:
            self.restart_kernel()

        self.log("Executing code: " + str(code)[:100] + "...")

        msg_id = self.client.execute(code)
        output = ""
        while True:
            try:
                msg = self.client.get_iopub_msg(timeout=1)
            except:
                continue

            if msg['parent_header'].get('msg_id') != msg_id:
                continue

            chunk = str(self.process_message(msg))
            if chunk.strip() != "":
                output += chunk
                self.send_output(chunk)

            if (msg['msg_type'] == 'status' and
                    msg['content']['execution_state'] == 'idle'):
                break

        return self.remove_ansi(output).strip()

    def restart_kernel(self) -> bool:
        """
        Restart the IPython kernel and reconnect.

        :return: True if the kernel was restarted successfully, False otherwise.
        """
        self.send_output("Restarting...")
        self.restart_container(self.get_container_name())
        if self.client is not None:
            self.client.stop_channels()
        else:
            self.client = BlockingKernelClient(connection_file=self.get_kernel_file_path())
        self.client.load_connection_file()
        self.client.start_channels()
        self.client.wait_for_ready()
        self.log("Connected to IPython kernel.")
        self.send_output("Restarted.")
        return True

    def send_output(self, output: str):
        """
        Send the output to the output.

        :param output: Output.
        :return: Output.
        """
        if self.signals is not None:
            self.signals.ipython_output.emit(output)

    def is_docker_installed(self) -> bool:
        """
        Check if Docker is installed

        :return: True if installed
        """
        try:
            if self.client is None:
                client = docker.from_env()
                client.ping()
            return True
        except DockerException:
            return False

    def run(self):
        """Run the IPython kernel. (debug console)"""
        try:
            self.init()

            # loop to send code to the kernel
            while True:
                code = input('>>> ')
                if code.strip() == 'exit':
                    break
                elif code.strip() == 'restart':
                    self.restart_kernel()
                    continue
                elif code.strip() in ['quit', 'exit', 'stop']:
                    self.stop_container(self.get_container_name())
                    break

                # send the execution request
                msg_id = self.client.execute(code)

                # get messages until execution is idle
                while True:
                    try:
                        msg = self.client.get_iopub_msg(timeout=1)
                    except:
                        continue

                    if msg['parent_header'].get('msg_id') != msg_id:
                        continue

                    # show output
                    output = str(self.process_message(msg)).strip()
                    if output:
                        print(output)

                    if (msg['msg_type'] == 'status' and
                            msg['content']['execution_state'] == 'idle'):
                        # execution is complete
                        break
        except KeyboardInterrupt:
            pass
        finally:
            if self.initialized:
                self.end()  # stop the client

    def remove_ansi_more(self, text):
        """
        Clean the text from ANSI escape sequences, carriage returns, and progress bars.

        :param text: Text to clean.
        :return: Cleaned text.
        """
        # Remove ANSI escape sequences
        ansi_escape = re.compile(r'''
            \x1B  # ESC
            (?:   # Start of sequence
                [@-Z\\-_]  # ESC [@ to ESC _ (7-bit C1 Control codes)
            |     # or
                \[  # ESC[
                [0-?]*  # Parameter bytes
                [ -/]*  # Intermediate bytes
                [@-~]   # Final byte
            )
        ''', re.VERBOSE)
        text = ansi_escape.sub('', text)

        # Split text into lines
        lines = text.split('\n')

        cleaned_lines = []
        for line in lines:
            # Handle carriage returns - keep only the text after the last '\r'
            if '\r' in line:
                line = line.split('\r')[-1]
            # Optionally, remove progress bar lines by detecting lines containing box-drawing characters
            # Skip the line if it contains box-drawing characters
            if re.search(r'[\u2500-\u257F]', line):
                continue  # Skip this line
            # Append the cleaned line
            cleaned_lines.append(line)
        # Reconstruct the text
        text = '\n'.join(cleaned_lines)
        return text

    def remove_ansi(self, text) -> str:
        """
        Clean the text from ANSI escape sequences.

        :param text: Text to clean.
        :return: Cleaned text.
        """
        ansi_escape = re.compile(
            r'''
            \x1B   # ESC
            (?:    # 7-bit C1 Fe
                [@-Z\\-_]
            |      # or 8-bit C1 Fe
                \[
                [0-?]*   # Parameter bytes
                [ -/]*   # Intermediate bytes
                [@-~]    # Final byte
            )
            ''',
            re.VERBOSE
        )
        return ansi_escape.sub('', text)

    def attach_signals(self, signals):
        """
        Attach signals

        :param signals: signals
        """
        self.signals = signals

    def log(self, msg):
        """
        Log the message.

        :param msg: Message to log.
        """
        print(msg)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict

from pygpt_net.item.attachment import AttachmentItem
from pygpt_net.item.ctx import CtxItem
from pygpt_net.utils import trans


class Files:
    def __init__(self, window=None):
        """
        Chat files/attachments controller

        :param window: Window instance
        """
        self.window = window

    def send(
            self,
            mode: str,
            ctx: CtxItem
    ):
        """
        Send attachments

        :param mode: mode
        :param ctx: CtxItem instance
        """
        # upload assistant attachments (only assistant mode here)
        attachments = self.upload(mode)  # current thread is already in global config
        if len(attachments) > 0:
            ctx.attachments = attachments
            self.window.update_status(trans('status.sending'))
            self.window.controller.chat.log("Uploaded attachments (Assistant): {}".format(len(attachments)))
        return attachments

    def upload(self, mode: str) -> Dict[str, AttachmentItem]:
        """
        Upload attachments

        :param mode: mode
        :return: uploaded attachments list
        """
        attachments_list = {}

        if mode == 'assistant':
            is_upload = False
            num_uploaded = 0
            try:
                # upload only new attachments (not uploaded yet to remote)
                attachments = self.window.core.attachments.get_all(mode)
                c = self.window.controller.assistant.files.count_upload(attachments)
                if c > 0:
                    is_upload = True
                    self.window.update_status(trans('status.uploading'))
                    num_uploaded = self.window.controller.assistant.files.upload(
                        mode,
                        attachments,
                    )
                    self.window.controller.files.uploaded_ids = self.window.core.attachments.get_ids(mode)
                    attachments_list = self.window.core.attachments.make_json_list(attachments)

                # show uploaded status
                if is_upload and num_uploaded > 0:
                    self.window.update_status(trans('status.uploaded'))

            except Exception as e:
                self.window.core.debug.log(e)
                self.window.ui.dialogs.alert(e)

        return attachments_list



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.19 02:00:00                  #
# ================================================== #

from PySide6.QtCore import Qt, QEvent
from PySide6.QtGui import QAction, QIcon, QKeySequence
from PySide6.QtWidgets import QTextEdit, QWidget, QVBoxLayout

from pygpt_net.core.tabs.tab import Tab
from pygpt_net.core.text.finder import Finder
from pygpt_net.ui.widget.element.labels import HelpLabel
from pygpt_net.utils import trans
import pygpt_net.icons_rc


class NotepadWidget(QWidget):
    def __init__(self, window=None):
        """
        Notepad

        :param window: main window
        """
        super(NotepadWidget, self).__init__(window)
        self.window = window
        self.id = 1  # assigned in setup
        self.textarea = NotepadOutput(self.window)
        self.window.ui.nodes['tip.output.tab.notepad'] = HelpLabel(trans('tip.output.tab.notepad'), self.window)
        self.opened = False
        self.tab = None

        layout = QVBoxLayout()
        layout.addWidget(self.textarea)
        layout.addWidget(self.window.ui.nodes['tip.output.tab.notepad'])
        layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(layout)

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab
        self.textarea.set_tab(tab)

    def scroll_to_bottom(self):
        """Scroll down"""
        self.textarea.verticalScrollBar().setValue(self.textarea.verticalScrollBar().maximum())

    def setText(self, text: str):
        """
        Set text

        :param text: Text
        """
        self.textarea.setText(text)
        self.textarea.on_update()

    def toPlainText(self) -> str:
        """
        Get plain text

        :return: Plain text
        """
        return self.textarea.toPlainText()

    def on_destroy(self):
        """On destroy"""
        # unregister finder from memory
        self.window.controller.finder.unset(self.textarea.finder)


class NotepadOutput(QTextEdit):
    def __init__(self, window=None):
        """
        Notepad output textarea

        :param window: main window
        """
        super(NotepadOutput, self).__init__(window)
        self.window = window
        self.finder = Finder(window, self)
        self.setAcceptRichText(False)
        self.setStyleSheet(self.window.controller.theme.style('font.chat.output'))
        self.value = self.window.core.config.data['font_size']
        self.max_font_size = 42
        self.min_font_size = 8
        self.id = 1  # assigned in setup
        self.textChanged.connect(self.text_changed)
        self.tab = None
        self.installEventFilter(self)

    def eventFilter(self, source, event):
        """
        Focus event filter

        :param source: source
        :param event: event
        """
        if event.type() == event.Type.FocusIn:
            if self.tab is not None:
                col_idx = self.tab.column_idx
                self.window.controller.ui.tabs.on_column_focus(col_idx)
        return super().eventFilter(source, event)

    def set_tab(self, tab: Tab):
        """
        Set tab

        :param tab: Tab
        """
        self.tab = tab

    def text_changed(self):
        """On text changed"""
        if not self.window.core.notepad.locked:
            self.window.controller.notepad.save(self.id)  # use notepad id
        self.finder.text_changed()

    def contextMenuEvent(self, event):
        """
        Context menu event

        :param event: Event
        """
        menu = self.createStandardContextMenu()
        selected_text = self.textCursor().selectedText()
        if selected_text:
            # plain text
            plain_text = self.textCursor().selection().toPlainText()

            # audio read
            action = QAction(QIcon(":/icons/volume.svg"), trans('text.context_menu.audio.read'), self)
            action.triggered.connect(self.audio_read_selection)
            menu.addAction(action)

            # copy to (without current notepad)
            excluded_id = "notepad_id_{}".format(self.id)
            copy_to_menu = self.window.ui.context_menu.get_copy_to_menu(self, selected_text, excluded=[excluded_id])
            menu.addMenu(copy_to_menu)

            # save as (selected)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_selection_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(plain_text))
            menu.addAction(action)
        else:
            # save as (all)
            action = QAction(QIcon(":/icons/save.svg"), trans('action.save_as'), self)
            action.triggered.connect(
                lambda: self.window.controller.chat.common.save_text(self.toPlainText()))
            menu.addAction(action)

        action = QAction(QIcon(":/icons/search.svg"), trans('text.context_menu.find'), self)
        action.triggered.connect(self.find_open)
        action.setShortcut(QKeySequence("Ctrl+F"))
        menu.addAction(action)

        menu.exec_(event.globalPos())

    def audio_read_selection(self):
        """
        Read selected text (audio)
        """
        self.window.controller.audio.read_text(self.textCursor().selectedText())

    def find_open(self):
        """Open find dialog"""
        self.window.controller.finder.open(self.finder)

    def on_update(self):
        """On content update"""
        self.finder.clear()  # clear finder

    def keyPressEvent(self, e):
        """
        Key press event

        :param e: Event
        """
        if e.key() == Qt.Key_F and e.modifiers() & Qt.ControlModifier:
            self.find_open()
        else:
            self.finder.clear(restore=True, to_end=False)
            super(NotepadOutput, self).keyPressEvent(e)

    def wheelEvent(self, event):
        """
        Wheel event: set font size

        :param event: Event
        """
        if event.modifiers() & Qt.ControlModifier:
            if event.angleDelta().y() > 0:
                if self.value < self.max_font_size:
                    self.value += 1
            else:
                if self.value > self.min_font_size:
                    self.value -= 1

            self.window.core.config.data['font_size'] = self.value
            self.window.core.config.save()
            option = self.window.controller.settings.editor.get_option('font_size')
            option['value'] = self.value
            self.window.controller.config.apply(
                parent_id='config', 
                key='font_size', 
                option=option,
            )
            self.window.controller.ui.update_font_size()
            event.accept()
        else:
            super(NotepadOutput, self).wheelEvent(event)

    def focusInEvent(self, e):
        """
        Focus in event

        :param e: focus event
        """
        self.window.controller.finder.focus_in(self.finder)
        super(NotepadOutput, self).focusInEvent(e)

    def focusOutEvent(self, e):
        """
        Focus out event

        :param e: focus event
        """
        super(NotepadOutput, self).focusOutEvent(e)
        self.window.controller.finder.focus_out(self.finder)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.21 20:00:00                  #
# ================================================== #

from pygpt_net.core.types import (
    MODE_AGENT,
)
from pygpt_net.utils import trans


class Common:
    def __init__(self, window=None):
        """
        Agent common controller

        :param window: Window instance
        """
        self.window = window

    def enable_auto_stop(self):
        """Enable auto stop (Legacy)"""
        self.window.core.config.set('agent.auto_stop', True)
        self.window.core.config.save()

    def disable_auto_stop(self):
        """Disable auto stop (Legacy)"""
        self.window.core.config.set('agent.auto_stop', False)
        self.window.core.config.save()

    def toggle_auto_stop(self, state: bool):
        """
        Toggle auto stop (Legacy)

        :param state: state of checkbox
        """
        if not state:
            self.disable_auto_stop()
        else:
            self.enable_auto_stop()

    def enable_continue(self):
        """Enable always continue (Legacy)"""
        self.window.core.config.set('agent.continue.always', True)
        self.window.core.config.save()

    def disable_continue(self):
        """Disable always continue (Legacy)"""
        self.window.core.config.set('agent.continue.always', False)
        self.window.core.config.save()

    def toggle_continue(self, state: bool):
        """
        Toggle always continue (Legacy)

        :param state: state of checkbox
        """
        if not state:
            self.disable_continue()
        else:
            self.enable_continue()

    def is_infinity_loop(self, mode: str) -> bool:
        """
        Check if infinity loop is active

        :param mode: current mode
        :return: True if infinity loop is enabled
        """
        # legacy
        if (mode == MODE_AGENT and self.window.core.config.get('agent.iterations') == 0) or \
            (self.window.controller.plugins.is_enabled("agent")
             and self.window.core.plugins.get_option("agent", "iterations") == 0):
            return True
        return False

    def display_infinity_loop_confirm(self):
        """Show infinity run confirm dialog"""
        self.window.ui.dialogs.confirm(
            type="agent.infinity.run",
            id=0,
            msg=trans("agent.infinity.confirm.content"),
        )

    def show_status(self):
        """Show agent status (Legacy)"""
        self.window.ui.nodes['status.agent'].setVisible(True)

    def hide_status(self):
        """Hide agent status (Legacy)"""
        self.window.ui.nodes['status.agent'].setVisible(False)

    def toggle_status(self):
        """Toggle agent status (Legacy)"""
        mode = self.window.core.config.get('mode')
        if mode in [MODE_AGENT] or self.window.controller.agent.legacy.is_inline():
            self.show_status()
        else:
            self.hide_status()

    def enable_loop(self):
        """Enable loop (Llama)"""
        self.window.core.config.set('agent.llama.loop.enabled', True)
        self.window.core.config.save()

    def disable_loop(self):
        """Disable loop (Llama)"""
        self.window.core.config.set('agent.llama.loop.enabled', False)
        self.window.core.config.save()

    def toggle_loop(self, state: bool):
        """
        Toggle loop (Llama)

        :param state: state of checkbox
        """
        if not state:
            self.disable_loop()
        else:
            self.enable_loop()


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import List, Dict

from pygpt_net.core.events import ControlEvent, AppEvent


class Actions:
    def __init__(self, window=None):
        """
        Actions core

        :param window: Window instance
        """
        self.window = window

    def get_access_choices(self) -> List[Dict[str, str]]:
        """
        Get app access events choices

        :return: choices
        """
        events_list = [
            ControlEvent.APP_STATUS,
            ControlEvent.APP_EXIT,
            ControlEvent.AUDIO_OUTPUT_ENABLE,
            ControlEvent.AUDIO_OUTPUT_DISABLE,
            ControlEvent.AUDIO_INPUT_ENABLE,
            ControlEvent.AUDIO_INPUT_DISABLE,
            ControlEvent.CAMERA_ENABLE,
            ControlEvent.CAMERA_DISABLE,
            ControlEvent.CAMERA_CAPTURE,
            ControlEvent.CTX_NEW,
            ControlEvent.CTX_PREV,
            ControlEvent.CTX_NEXT,
            ControlEvent.CTX_LAST,
            ControlEvent.CTX_INPUT_FOCUS,
            ControlEvent.CTX_INPUT_SEND,
            ControlEvent.CTX_INPUT_CLEAR,
            ControlEvent.CTX_STOP,
            ControlEvent.CTX_ATTACHMENTS_CLEAR,
            ControlEvent.MODE_CHAT,
            ControlEvent.MODE_LLAMA_INDEX,
            ControlEvent.MODE_NEXT,
            ControlEvent.MODE_PREV,
            ControlEvent.MODEL_NEXT,
            ControlEvent.MODEL_PREV,
            ControlEvent.PRESET_NEXT,
            ControlEvent.PRESET_PREV,
            ControlEvent.TAB_CHAT,
            ControlEvent.TAB_CALENDAR,
            ControlEvent.TAB_DRAW,
            ControlEvent.TAB_FILES,
            ControlEvent.TAB_NOTEPAD,
            ControlEvent.TAB_NEXT,
            ControlEvent.TAB_PREV,
            ControlEvent.VOICE_MESSAGE_START,
            ControlEvent.VOICE_MESSAGE_STOP,
            ControlEvent.VOICE_MESSAGE_TOGGLE,
            ControlEvent.VOICE_COMMAND_START,
            ControlEvent.VOICE_COMMAND_STOP,
            ControlEvent.VOICE_COMMAND_TOGGLE,
        ]
        # build choices
        choices = []
        for event in events_list:
            choices.append({event: f"event.control.{event}"})
        return choices

    def get_speech_synthesis_choices(self) -> List[Dict[str, str]]:
        """
        Get app access events choices

        :return: choices
        """
        events_list_control = [
            ControlEvent.CALENDAR_ADD,
            ControlEvent.CALENDAR_CLEAR,
            ControlEvent.CTX_RENAME,
            ControlEvent.CTX_SEARCH_STRING,
            ControlEvent.CTX_SEARCH_CLEAR,
            ControlEvent.INPUT_APPEND,
            ControlEvent.NOTE_ADD,
            ControlEvent.NOTEPAD_CLEAR,
        ]
        events_list_app = [
            AppEvent.APP_STARTED,
            AppEvent.CTX_CREATED,
            AppEvent.CTX_END,
            AppEvent.CTX_SELECTED,
            AppEvent.CTX_ATTACHMENTS_CLEAR,
            AppEvent.CAMERA_ENABLED,
            AppEvent.CAMERA_DISABLED,
            AppEvent.CAMERA_CAPTURED,
            AppEvent.INPUT_ERROR,
            AppEvent.INPUT_SENT,
            AppEvent.INPUT_STOPPED,
            AppEvent.MODE_SELECTED,
            AppEvent.MODEL_SELECTED,
            AppEvent.PRESET_SELECTED,
            AppEvent.TAB_SELECTED,
        ]
        # build choices
        choices = []
        for event in events_list_control:
            choices.append({event: f"event.audio.{event}"})
        for event in events_list_app:
            choices.append({event: f"event.audio.{event}"})
        return choices

    def get_voice_control_choices(self) -> List[Dict[str, str]]:
        """
        Get voice control choices

        :return: choices
        """
        actions = self.window.core.access.voice.get_commands().items()
        # build choices
        choices = []
        for k, v in actions:
            choices.append({k: v})
        return choices


# https://github.com/run-llama/llama_index/blob/main/llama-index-integrations/readers/llama-index-readers-youtube-transcript/llama_index/readers/youtube_transcript/base.py
"""Simple Reader that reads transcript of youtube video."""
import re
from typing import Any, List, Optional

from youtube_transcript_api import YouTubeTranscriptApi

from llama_index.core.readers.base import BasePydanticReader
from llama_index.core.schema import Document

from .utils import YOUTUBE_URL_PATTERNS


class YoutubeTranscriptReader(BasePydanticReader):
    """Youtube Transcript reader."""

    is_remote: bool = True
    languages: List[str] = ["en"]

    @classmethod
    def class_name(cls) -> str:
        """Get the name identifier of the class."""
        return "YoutubeTranscriptReader"

    def load_data(
        self,
        ytlinks: List[str],
        languages: Optional[List[str]] = ["en"],
        **load_kwargs: Any,
    ) -> List[Document]:
        """Load data from the input directory.

        Args:
            pages (List[str]): List of youtube links \
                for which transcripts are to be read.

        """
        languages = self.languages
        if not languages:
            languages = ["en"]

        results = []
        for link in ytlinks:
            video_id = self._extract_video_id(link)
            if not video_id:
                raise ValueError(
                    f"Supplied url {link} is not a supported youtube URL."
                    "Supported formats include:"
                    "  youtube.com/watch?v=\\{video_id\\} "
                    "(with or without 'www.')\n"
                    "  youtube.com/embed?v=\\{video_id\\} "
                    "(with or without 'www.')\n"
                    "  youtu.be/{video_id\\} (never includes www subdomain)"
                )
            transcript_chunks = YouTubeTranscriptApi.get_transcript(
                video_id, languages=languages
            )
            chunk_text = [chunk["text"] for chunk in transcript_chunks]
            transcript = "\n".join(chunk_text)
            results.append(
                Document(
                    text=transcript, id_=video_id, metadata={"video_id": video_id}
                )
            )
        return results

    @staticmethod
    def _extract_video_id(yt_link) -> Optional[str]:
        for pattern in YOUTUBE_URL_PATTERNS:
            match = re.search(pattern, yt_link)
            if match:
                return match.group(1)

        # return None if no match is found
        return None


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.14 01:00:00                  #
# ================================================== #

from PySide6.QtCore import QObject, Signal, QRunnable, Slot


class IndexWorkerSignals(QObject):
    finished = Signal(str, object, object, bool)  # idx, result, errors, silent mode
    error = Signal(object)


class IndexWorker(QObject, QRunnable):
    def __init__(self, *args, **kwargs):
        QObject.__init__(self)
        QRunnable.__init__(self)
        self.signals = IndexWorkerSignals()
        self.window = None
        self.content = None
        self.loader = None
        self.params = None
        self.config = None
        self.replace = None
        self.recursive = None
        self.from_ts = 0
        self.idx = None
        self.type = None
        self.silent = False

    @Slot()
    def run(self):
        """Indexer thread"""
        try:
            result = {}
            errors = []

            # log
            self.log("Indexing data...")
            self.log("Idx: {}, type: {}, content: {}, from_ts: {}".format(
                self.idx,
                self.type,
                self.content,
                self.from_ts,
            ))

            # execute indexing
            if self.type == "file":
                result, errors = self.window.core.idx.index_files(
                    self.idx,
                    self.content,
                    self.replace,
                    self.recursive,
                )
            elif self.type == "files":
                result = {}
                errors = []
                for path in self.content:
                    r, e = self.window.core.idx.index_files(
                        self.idx,
                        path,
                        self.replace,
                        self.recursive,
                    )
                    result.update(r)
                    errors.extend(e)
            elif self.type == "db_meta":
                result, errors = self.window.core.idx.index_db_by_meta_id(
                    self.idx,
                    self.content,
                    self.from_ts,
                )
            elif self.type == "db_current":
                result, errors = self.window.core.idx.index_db_from_updated_ts(
                    self.idx,
                    self.content,
                )
            elif self.type == "web":
                result, errors = self.window.core.idx.index_web(
                    idx=self.idx,
                    type=self.loader,
                    params=self.params,
                    config=self.config,
                    replace=self.replace,
                )

            self.log("Finished indexing.")
            self.signals.finished.emit(
                self.idx,
                result,
                errors,
                self.silent,
            )
        except Exception as e:
            self.window.core.debug.error(e)
            self.signals.error.emit(e)

    def log(self, msg: str):
        """
        Log info message

        :param msg: message
        """
        is_log = False
        if self.window.core.config.has("log.llama") \
                and self.window.core.config.get("log.llama"):
            is_log = True
        self.window.core.debug.info(msg, not is_log)
        if is_log:
            print("[LLAMA-INDEX] {}".format(msg))
        self.window.idx_logger_message.emit(msg)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import datetime
import os.path
from typing import List, Dict, Any

from llama_index.core.schema import Document


class Metadata:
    def __init__(self, window=None):
        """
        Metadata core

        :param window: Window instance
        """
        self.window = window

    def append_file_metadata(
            self,
            docs: List[Document],
            path: str
    ):
        """
        Append custom meta to documents (file)

        :param docs: list of documents
        :param path: path to file
        """
        metas = self.window.core.config.get("llama.idx.custom_meta")
        ext = str(os.path.splitext(path)[1][1:]).lower()
        data_dir = self.window.core.config.get_user_dir("data")
        if (len(docs) > 0
                and metas is not None
                and isinstance(metas, list)
                and len(metas) > 0):

            for doc in docs:
                meta = doc.metadata
                for item in metas:
                    extensions = item["extensions"].replace(" ", "").lower().split(",")

                    # * means all extensions
                    if ext in extensions or "*" in extensions:
                        key = item["key"]
                        value = item["value"]

                        # remove key if value is empty
                        if str(value).strip() == "" and key in meta:
                            del meta[key]
                            continue

                        # append or replace custom meta
                        try:
                            relative_path = os.path.relpath(path, data_dir)
                            relative_dir = os.path.relpath(os.path.dirname(path), data_dir)
                            # if file is outside of data directory, strip relative path, and use absolute path
                            if relative_path.startswith(".."):
                                relative_path = os.path.basename(path)
                                relative_dir = os.path.dirname(path)
                            meta[key] = value.format(
                                path=path,
                                relative_path=relative_path,
                                relative_dir=relative_dir,
                                filename=os.path.basename(path),
                                dirname=os.path.dirname(path),
                                ext=ext,
                                size=os.path.getsize(path),
                                mtime=os.path.getmtime(path),
                                timestamp=int(datetime.datetime.now().timestamp()),
                                date=datetime.datetime.now().strftime("%Y-%m-%d"),
                                time=datetime.datetime.now().strftime("%H:%M:%S"),
                                date_time=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                            )
                        except Exception as e:
                            print("Error while appending custom meta: " + str(e))
                            self.window.core.debug.log(e)
                doc.metadata = meta

    def append_web_metadata(
            self,
            docs: List[Document],
            type: str,
            args: Dict[str, Any]
    ):
        """
        Append custom meta to documents (web)

        :param docs: list of documents
        :param type: type of URL (webpage, feed, etc.)
        :param args: loader arguments
        """
        metas = self.window.core.config.get("llama.idx.custom_meta.web")
        if (len(docs) > 0
                and metas is not None
                and isinstance(metas, list)
                and len(metas) > 0):

            for doc in docs:
                meta = doc.metadata
                for item in metas:
                    item_type = item["loader"].replace("web_", "")
                    if item_type != type:
                        continue

                    key = item["key"]
                    value = item["value"]

                    # remove key if value is empty
                    if str(value).strip() == "" and key in meta:
                        del meta[key]
                        continue

                    # append or replace custom meta
                    try:
                        meta[key] = value.format(
                            timestamp=int(datetime.datetime.now().timestamp()),
                            date=datetime.datetime.now().strftime("%Y-%m-%d"),
                            time=datetime.datetime.now().strftime("%H:%M:%S"),
                            date_time=datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
                            **args
                        )
                    except Exception as e:
                        print("Error while appending custom meta (web): " + str(e))
                        self.window.core.debug.log(e)
                doc.metadata = meta




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

import datetime
from typing import Optional, Dict

from packaging.version import Version

from pygpt_net.item.calendar_note import CalendarNoteItem
from pygpt_net.provider.core.calendar.db_sqlite import DbSqliteProvider


class Calendar:
    def __init__(self, window=None):
        """
        Calendar core

        :param window: Window instance
        """
        self.window = window
        self.provider = DbSqliteProvider(window)
        self.items = {}

    def install(self):
        """Install provider data"""
        self.provider.install()

    def patch(self, app_version: Version):
        """
        Patch provider data

        :param app_version: app version
        """
        self.provider.patch(app_version)

    def get_by_date(
            self,
            year: int,
            month: int,
            day: int
    ) -> Optional[CalendarNoteItem]:
        """
        Get note by date

        :param year: year
        :param month: month
        :param day: day
        :return: notepad instance
        """
        # convert to format: YYYY-MM-DD:
        dt_key = datetime.datetime(year, month, day).strftime("%Y-%m-%d")
        if dt_key in self.items:
            return self.items[dt_key]
        return None

    def get_all(self) -> Dict[str, CalendarNoteItem]:
        """
        Get all notes

        :return: notepads dict
        """
        return self.items

    def build(self) -> CalendarNoteItem:
        """
        Build note

        :return: CalendarNoteItem instance
        """
        item = CalendarNoteItem()
        return item

    def add(self, note: CalendarNoteItem) -> bool:
        """
        Add note

        :param note: CalendarNoteItem instance
        :return: True if success
        """
        self.provider.create(note)
        dt_key = datetime.datetime(note.year, note.month, note.day).strftime("%Y-%m-%d")
        self.items[dt_key] = note
        self.save(note.year, note.month, note.day)
        return True

    def update(self, note: CalendarNoteItem) -> bool:
        """
        Update and save note

        :param note: CalendarNoteItem instance
        :return: True if success
        """
        # convert to format: YYYY-MM-DD:
        dt_key = datetime.datetime(note.year, note.month, note.day).strftime("%Y-%m-%d")
        if dt_key not in self.items:
            return False
        note.updated_at = datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S")
        self.items[dt_key] = note
        self.save(note.year, note.month, note.day)
        return True

    def load(
            self,
            year: int,
            month: int,
            day: int
    ):
        """
        Load note by idx

        :param year: year
        :param month: month
        :param day: day
        """
        # convert to format: YYYY-MM-DD:
        dt_key = datetime.datetime(year, month, day).strftime("%Y-%m-%d")
        self.items[dt_key] = self.provider.load(year, month, day)

    def load_note(
            self,
            year: int,
            month: int,
            day: int
    ) -> str:
        """
        Load note by idx

        :param year: year
        :param month: month
        :param day: day
        :return: note content
        """
        note = self.provider.load(year, month, day)
        if note is not None:
            return str(note.content)
        return ""

    def append_to_note(
            self,
            year: int,
            month: int,
            day: int,
            text: str
    ) -> bool:
        """
        Append to note

        :param year: year
        :param month: month
        :param day: day
        :param text: text
        :return: True if success
        """
        current = self.provider.load(year, month, day)
        if current is None:
            current = self.build()
        if current.content is not None and current.content != "":
            current.content = current.content + "\n"
        if current.content is None:
            current.content = ""
        current.year = year
        current.month = month
        current.day = day
        current.content += text
        self.provider.save(current)
        self.load(year, month, day)
        return True

    def update_note(
            self,
            year: int,
            month: int,
            day: int,
            text: str
    ) -> bool:
        """
        Append to note

        :param year: year
        :param month: month
        :param day: day
        :param text: text
        :return: True if success
        """
        current = self.provider.load(year, month, day)
        if current is None:
            current = self.build()
            current.year = year
            current.month = month
            current.day = day
        current.content = text
        self.provider.save(current)
        self.load(year, month, day)
        return True

    def remove_note(
            self,
            year: int,
            month: int,
            day: int
    ) -> bool:
        """
        Remove note by date

        :param year: year
        :param month: month
        :param day: day
        :return: True if removed
        """
        self.provider.remove(year, month, day)
        # convert to format: YYYY-MM-DD:
        dt_key = datetime.datetime(year, month, day).strftime("%Y-%m-%d")
        if dt_key in self.items:
            del self.items[dt_key]
        return True

    def load_all(self):
        """Load all notes"""
        self.items = self.provider.load_all()

    def load_by_month(self, year: int, month: int):
        """
        Load notes by month

        :param year: year
        :param month: month
        """
        self.items = self.provider.load_by_month(year, month)

    def get_notes_existence_by_day(
            self,
            year: int,
            month: int
    ) -> Dict[str, Dict[int, int]]:
        """
        Get notes existence by day

        :param year: year
        :param month: month
        :return: dict
        """
        return self.provider.get_notes_existence_by_day(year, month)

    def save(
            self,
            year: int,
            month: int,
            day: int
    ) -> bool:
        """
        Save note by date

        :param year: year
        :param month: month
        :param day: day
        :return: True if saved, False if not
        """
        # convert to format: YYYY-MM-DD:
        dt_key = datetime.datetime(year, month, day).strftime("%Y-%m-%d")
        if dt_key not in self.items:
            return False

        self.provider.save(self.items[dt_key])
        return False

    def save_all(self):
        """Save all notes"""
        self.provider.save_all(self.items)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.01.18 03:00:00                  #
# ================================================== #

from typing import Any

from PySide6.QtCore import Slot

from pygpt_net.core.types import MODE_AUDIO
from pygpt_net.plugin.base.plugin import BasePlugin
from pygpt_net.provider.audio_output.base import BaseProvider
from pygpt_net.core.events import Event
from pygpt_net.item.ctx import CtxItem

from .config import Config
from .worker import Worker


class Plugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        super(Plugin, self).__init__(*args, **kwargs)
        self.id = "audio_output"
        self.name = "Audio Output"
        self.type = ['audio.output']
        self.description = "Enables audio/voice output (speech synthesis)"
        self.prefix = "Audio Output"
        self.input_text = None
        self.playback = None
        self.order = 1
        self.use_locale = True
        self.output_file = "output.mp3"
        self.config = Config(self)
        self.worker = None

    def init_options(self):
        """Initialize options"""
        self.config.from_defaults(self)
        self.init_provider()

    def init_provider(self):
        """Initialize provider options"""
        providers = self.get_providers()
        for id in providers:
            providers[id].init(self)

    def get_providers(self) -> dict:
        """
        Get audio output providers

        :return: providers dict
        """
        return self.window.core.audio.get_providers("output")

    def get_provider_options(self) -> list:
        """Get provider options"""
        options = []
        providers = self.get_providers()
        for id in providers:
            options.append({id: providers[id].name})
        return options

    def init_tabs(self) -> dict:
        """
        Initialize provider tabs

        :return: dict of tabs
        """
        tabs = {}
        tabs["general"] = "General"
        providers = self.get_providers()
        for id in providers:
            tabs[id] = providers[id].name
        return tabs

    def setup_ui(self):
        """
        Setup UI
        """
        pass

    def attach(self, window):
        """
        Attach window

        :param window: Window instance
        """
        self.window = window

        # register provider tabs
        self.tabs = self.init_tabs()

        # register options
        self.init_options()

    def get_provider(self) -> BaseProvider:
        """
        Get audio output provider

        :return: provider instance
        """
        current = self.get_option_value("provider")
        providers = self.get_providers()
        if current not in providers:
            raise Exception("Provider '{}' not found!".format(current))
        return providers[current]

    def handle(self, event: Event, *args, **kwargs):
        """
        Handle dispatched event

        :param event: event object
        :param args: args
        :param kwargs: kwargs
        """
        name = event.name
        data = event.data
        ctx = event.ctx
        mode = self.window.core.config.get("mode")

        if name == Event.INPUT_BEFORE:
            self.on_input_before(data['value'])

        elif name == Event.CTX_AFTER:
            if mode == MODE_AUDIO:
                return  # skip if audio mode
            self.stop_audio()
            self.on_generate(ctx, event)

        elif name == Event.AUDIO_READ_TEXT:
            self.stop_audio()
            self.on_generate(ctx, event)

        elif name == Event.AUDIO_PLAYBACK:
            self.stop_audio()
            self.on_playback(ctx, event)

        elif name == Event.AUDIO_OUTPUT_STOP:
            self.stop_audio()

    def on_input_before(self, text: str):
        """
        Event: INPUT_BEFORE

        :param text: text to read
        """
        self.input_text = text

    def on_generate(self, ctx: CtxItem, event: Event):
        """
        Events: CTX_AFTER, AUDIO_READ_TEXT

        :param ctx: CtxItem
        :param event: Event
        """
        # check if provider is configured
        if not self.get_provider().is_configured():
            msg = self.get_provider().get_config_message()
            self.window.ui.dialogs.alert(msg)
            return

        name = event.name
        text = ctx.output
        cache_file = None
        if event.data is not None and isinstance(event.data, dict) and "cache_file" in event.data:
            cache_file = event.data["cache_file"]

        # check for audio read allowed. Prevents reading audio in commands, results, etc.
        if name == Event.CTX_AFTER:
            if not ctx.audio_read_allowed():
                return  # abort if audio read is not allowed (commands, results, etc.)

        try:
            if text is not None and len(text) > 0:
                self.stop_audio()

                worker = Worker()
                worker.from_defaults(self)
                worker.ctx = ctx
                worker.event = name
                worker.cache_file = cache_file
                worker.text = self.window.core.audio.clean_text(text)
                worker.mode = "generate"

                # signals
                worker.signals.playback.connect(self.handle_playback)
                worker.signals.error_playback.connect(self.handle_playback_error)
                worker.signals.stop.connect(self.handle_stop)
                worker.signals.volume_changed.connect(self.handle_volume)

                worker.run_async()
                self.worker = worker

                # only for manual reading
                if name == Event.AUDIO_READ_TEXT:
                    self.window.controller.audio.on_begin(self.worker.text)

        except Exception as e:
            self.error(e)

    def on_playback(self, ctx: CtxItem, event: Event):
        """
        Events: AUDIO_PLAYBACK

        :param ctx: CtxItem
        :param event: Event
        """
        try:
            self.stop_audio()

            worker = Worker()
            worker.from_defaults(self)
            worker.audio_file = event.data["audio_file"]
            worker.mode = "playback"

            # signals
            worker.signals.playback.connect(self.handle_playback)
            worker.signals.error_playback.connect(self.handle_playback_error)
            worker.signals.stop.connect(self.handle_stop)
            worker.signals.volume_changed.connect(self.handle_volume)

            worker.run_async()
            self.worker = worker

        except Exception as e:
            self.error(e)

    def destroy(self):
        """Destroy thread"""
        pass

    def set_status(self, status: str):
        """
        Set status

        :param status: status text
        """
        self.window.ui.plugin_addon['audio.output'].set_status(status)

    def show_stop_button(self):
        """Show stop button"""
        self.window.ui.plugin_addon['audio.output'].stop.setVisible(True)

    def hide_stop_button(self):
        """Hide stop button"""
        self.window.ui.plugin_addon['audio.output'].stop.setVisible(False)

    def stop_speak(self):
        """Stop speaking"""
        self.window.ui.plugin_addon['audio.output'].stop.setVisible(False)
        self.window.ui.plugin_addon['audio.output'].set_status('Stopped')
        self.window.ui.plugin_addon['audio.output'].stop_audio()

    def stop_audio(self):
        """
        Event: AUDIO_OUTPUT_STOP

        Stop playing the audio
        """
        if self.worker is not None:
            self.worker.stop()
        self.handle_volume(0.0)

    @Slot(object)
    def handle_playback_error(self, err: Any):
        """
        Send error message to logger and alert dialog

        :param err: error message
        """
        self.error(err)
        if self.window.core.platforms.is_snap():
            self.window.ui.dialogs.open(
                'snap_audio_output',
                width=400,
                height=200
            )

    @Slot(str)
    def handle_playback(self, event: str):
        """
        Handle thread playback object

        :param event: event name
        """
        self.window.controller.audio.on_play(event)

    @Slot()
    def handle_stop(self):
        """Handle thread playback stop"""
        self.stop_audio()

    @Slot(float)
    def handle_volume(self, volume: float):
        """
        Handle thread playback volume

        :param volume: volume level
        """
        self.window.ui.plugin_addon['audio.output.bar'].setLevel(volume)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.11.23 21:00:00                  #
# ================================================== #

import os
import shutil

from uuid import uuid4

class Packer:
    def __init__(self, window=None):
        """
        Archive packer/unpacker

        :param window: Window instance
        """
        self.window = window
        self.zip_ext = [".zip"]
        self.tar_ext = [".tar", ".gz", ".bz2"]

    def is_archive(self, path: str) -> bool:
        """
        Check if path is archive

        :param path: Path to file
        :return: True if archive
        """
        ext = os.path.splitext(path)[1]
        return ext in self.zip_ext or ext in self.tar_ext

    def unpack(self, path: str) -> str:
        """
        Unpack archive

        :param path: path to archive
        :return: path to unpacked directory
        """
        uuid = str(uuid4())
        tmp_dir = os.path.join(self.window.core.config.get_user_dir("tmp"), uuid)
        os.makedirs(tmp_dir, exist_ok=True)
        ext = os.path.splitext(path)[1]
        if ext in self.zip_ext:
            return self.unpack_zip(path, tmp_dir)
        elif ext in self.tar_ext:
            return self.unpack_tar(path, tmp_dir)

    def unpack_zip(self, path: str, dir: str) -> str:
        """
        Unpack archive to temporary directory

        :param path: path to archive
        :param dir: temporary directory
        :return: path to unpacked directory
        """
        import zipfile
        with zipfile.ZipFile(path, 'r') as zip_ref:
            zip_ref.extractall(dir)
        return dir

    def unpack_tar(self, path: str, dir: str) -> str:
        """
        Unpack archive to temporary directory

        :param path: path to archive
        :param dir: temporary directory
        :return: path to unpacked directory
        """
        import tarfile
        with tarfile.open(path, 'r') as tar_ref:
            tar_ref.extractall(dir)
        return dir

    def remove_tmp(self, path: str):
        """
        Remove temporary directory

        :param path: path to directory
        """
        if os.path.exists(path) and os.path.isdir(path):
            shutil.rmtree(path)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2024.12.14 08:00:00                  #
# ================================================== #

from typing import Dict, Any

from pygpt_net.item.ctx import CtxItem


class AudioContext:
    def __init__(self, **kwargs):
        """
        Audio context

        :param kwargs: keyword arguments
        """
        self.data = kwargs.get("ctx", None)
        self.prev_id = kwargs.get("prev_id", None)

    def to_dict(self) -> Dict:
        """
        Return as dictionary

        :return: dictionary
        """
        data = {
            "data": self.data,
            "prev_id": self.prev_id
        }
        # sort by keys
        data = dict(sorted(data.items(), key=lambda item: item[0]))
        return data